from abc import abstractmethod
from sqlalchemy.orm import relationship

from investing_algorithm_framework.core.models.model_extension import \
    SQLAlchemyModelExtension
from investing_algorithm_framework.core.models import db


class AssetPrice:

    @abstractmethod
    def get_target_symbol(self):
        pass

    @abstractmethod
    def get_trading_symbol(self):
        pass

    @abstractmethod
    def get_price(self):
        pass

    @abstractmethod
    def get_datetime(self):
        pass

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"


class SQLLiteAssetPrice(AssetPrice, db.Model, SQLAlchemyModelExtension):

    __tablename__ = "asset_prices"

    id = db.Column(db.Integer, primary_key=True)
    target_symbol = db.Column(db.String, nullable=False)
    trading_symbol = db.Column(db.String, nullable=False)
    price = db.Column(db.Float, nullable=False)
    datetime = db.Column(db.DateTime, nullable=False)

    # Relationships
    asset_price_history_id = db.Column(
        db.Integer, db.ForeignKey('asset_price_histories.id')
    )
    asset_price_history = relationship(
        "SQLLiteAssetPriceHistory", back_populates="prices"
    )

    def __init__(self, target_symbol, trading_symbol, price, datetime):
        self.target_symbol = target_symbol
        self.trading_symbol = trading_symbol
        self.price = price
        self.datetime = datetime

    def get_target_symbol(self):
        return self.target_symbol

    def get_trading_symbol(self):
        return self.trading_symbol

    def get_price(self):
        return self.price

    def get_datetime(self):
        return self.datetime

    def __repr__(self):
        return self.repr(
            target_symbol=self.target_symbol,
            trading_symbol=self.trading_symbol,
            price=self.price,
            datetime=self.datetime
        )



import hashlib
import hmac


# Function to create a sha256 signature
def create_sha256_signature(secret_key: str, params: str):

    # Encode the secret key
    byte_secret_key = bytearray()
    byte_secret_key.extend(map(ord, secret_key))

    # Encode the params
    encoded_params = params.encode()

    # Create signature
    return hmac.new(byte_secret_key, encoded_params, hashlib.sha256)\
        .hexdigest()



import os
import logging
import logging.config
import marshmallow.exceptions as marshmallow_exceptions
from typing import Dict, List
from flask import Flask, jsonify
from flask_cors import CORS
from werkzeug.exceptions import HTTPException

from investing_algorithm_framework.configuration import ConfigValidator
from investing_algorithm_framework.configuration.constants import \
    DATABASE_CONFIG, DATABASE_DIRECTORY_PATH, DATABASE_NAME, LOG_LEVEL, \
    SQLALCHEMY_DATABASE_URI, RESOURCES_DIRECTORY
from investing_algorithm_framework.exceptions import ApiException

logger = logging.getLogger(__name__)

DEFAULT_DATABASE_NAME = "database"


def create_app(config_object=None) -> Flask:
    """
    Function to create a Flask app. The app will be based on the \
    given configuration from the client.
    """

    app = Flask(__name__.split('.')[0])
    CORS(app, supports_credentials=True)
    app.url_map.strict_slashes = False

    # Register error handler
    register_error_handlers(app)

    return app


def setup_config(flask_app, config_object):

    # Validate the configuration
    ConfigValidator.validate(config_object)

    for attribute_key in dir(config_object):
        if attribute_key.isupper():
            flask_app.config[attribute_key] = \
                getattr(config_object, attribute_key)


def setup_database(config_object):

    if DATABASE_CONFIG not in config_object:
        database_path = os.path.join(
            config_object[RESOURCES_DIRECTORY],
            '{}.sqlite3'.format(DEFAULT_DATABASE_NAME)
        )
        config_object[SQLALCHEMY_DATABASE_URI] = database_path
    else:
        database_name = DEFAULT_DATABASE_NAME
        database_directory_path = config_object.get(RESOURCES_DIRECTORY)

        if DATABASE_NAME in config_object.get(DATABASE_CONFIG):
            config_database_name = config_object.get(DATABASE_CONFIG)\
                .get(DATABASE_NAME)

            if config_database_name is not None:
                database_name = config_database_name

        if DATABASE_DIRECTORY_PATH in config_object.get(DATABASE_CONFIG):
            config_database_directory_path = config_object.get(DATABASE_CONFIG)\
                .get(DATABASE_DIRECTORY_PATH)

            if config_database_directory_path is not None:
                database_directory_path = config_database_directory_path

        config_object[DATABASE_CONFIG][DATABASE_DIRECTORY_PATH] = \
            database_directory_path
        config_object[DATABASE_CONFIG][DATABASE_NAME] = database_name

        database_path = os.path.join(
            config_object[DATABASE_CONFIG][DATABASE_DIRECTORY_PATH],
            f'{config_object[DATABASE_CONFIG][DATABASE_NAME]}.sqlite3'
        )

    config_object[SQLALCHEMY_DATABASE_URI] = 'sqlite:////{}'.format(database_path)

    # Create the database if it not exist
    if not os.path.isfile(database_path):
        open(database_path, 'w').close()


def setup_logging(log_level):
    pass
    # logging_config = {
    #     'version': 1,
    #     'disable_existing_loggers': True,
    #     'formatters': {
    #         'standard': {
    #             'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
    #         },
    #     },
    #     'handlers': {
    #         'console': {
    #             'level': 'INFO',
    #             'formatter': 'standard',
    #             'class': 'logging.StreamHandler',
    #             'stream': 'ext://sys.stdout',  # Default is stderr
    #         },
    #     },
    #     'loggers': {
    #         '': {  # root logger
    #             'handlers': ['console'],
    #             'level': "DEBUG",
    #             'propagate': False
    #         },
    #         'app': {
    #             'handlers': ['console'],
    #             'level': "DEBUG",
    #             'propagate': False
    #         },
    #     }
    # }
    #
    # logging.config.dictConfig(logging_config)


def register_error_handlers(app) -> None:
    """
    Function that will register all the specified error handlers for the app
    """

    def create_error_response(error_message, status_code: int = 400):

        # Remove the default 404 not found message if it exists
        if isinstance(error_message, str):
            error_message = error_message.replace("404 Not Found: ", '')

        response = jsonify({"error_message": error_message})
        response.status_code = status_code
        return response

    def format_marshmallow_validation_error(errors: Dict):
        errors_message = {}

        for key in errors:

            if isinstance(errors[key], Dict):
                errors_message[key] = \
                    format_marshmallow_validation_error(errors[key])

            if isinstance(errors[key], List):
                errors_message[key] = errors[key][0].lower()
        return errors_message

    def error_handler(error):
        logger.error("exception of type {} occurred".format(type(error)))
        logger.exception(error)

        if isinstance(error, HTTPException):
            return create_error_response(str(error), error.code)
        elif isinstance(error, ApiException):
            return create_error_response(
                error.error_message, error.status_code
            )
        elif isinstance(error, marshmallow_exceptions.ValidationError):
            error_message = format_marshmallow_validation_error(error.messages)
            return create_error_response(error_message)
        else:
            # Internal error happened that was unknown
            return "Internal server error", 500

    app.errorhandler(Exception)(error_handler)



from investing_algorithm_framework.core.models import TimeFrame


class AssetPricesQueue:
    """
    Asset prices queue for logical order of asset prices based on
    their date times. This data structure is useful to link
    asset prices to snapshots based on their date of creation.
    """

    def __init__(self, asset_prices_array, time_frame):
        self.asset_prices = []

        self.intervals = TimeFrame.from_value(time_frame).intervals

        for zipped_asset_prices in zip(*asset_prices_array):
            self.asset_prices.insert(0, zipped_asset_prices)

        self.size = len(self.intervals)
        self.front = self.size - 1

    def pop(self):

        if len(self.intervals) == 0:
            return None, []
        else:
            # Get the first item of the queue
            asset_prices = []

            if len(self.asset_prices) != 0:
                asset_prices = self.asset_prices.pop()

            interval = self.intervals.pop()

            self.size = len(self.intervals)
            self.front = self.size - 1

            return interval, asset_prices

    def peek(self):
        """
        Method to get the next snapshot in the queue without
        popping it from the stack.

        Returns None if the stack size is smaller than 1.
        """

        asset_prices = []

        if len(self.intervals) == 0:
            return None, None

        if len(self.asset_prices) >= 1:
            asset_prices = self.asset_prices[self.front]

        return self.intervals[self.front], asset_prices

    def empty(self):
        return len(self.intervals) == 0




from enum import Enum


class TradingDataTypes(Enum):
    TICKER = 'TICKER'
    ORDER_BOOK = 'ORDER_BOOK'
    RAW = "RAW"

    @staticmethod
    def from_value(value):

        if isinstance(value, TradingDataTypes):
            for trading_data_type in TradingDataTypes:

                if value == trading_data_type:
                    return trading_data_type

        elif isinstance(value, str):
            return TradingDataTypes.from_string(value)

        raise ValueError(
            "Could not convert value to trading data type"
        )

    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):
            for order_type in TradingDataTypes:

                if value.upper() == order_type.value:
                    return order_type

        raise ValueError(
            "Could not convert value to trading data type"
        )

    def equals(self, other):

        if other is None:
            return False

        if isinstance(other, Enum):
            return self.value == other.value

        else:
            return TradingDataTypes.from_string(other) == self



class SnapShotQueue:
    """
    Snapshot queue for logical order of snapshots. This data structure is
    useful to calculate the active time frame of a specific snapshot.

    e.g. snapshot is active from start time (snapshot.created_at) until
    end time (snapshot_peek.created_at)
    """

    def __init__(self, snapshots):
        self.queue = []

        for snapshot in snapshots:
            self.queue.append(snapshot)

        self.size = len(snapshots)
        self.front = self.size - 1

    def pop(self):

        if len(self.queue) == 0:
            return None
        else:
            self.size -= 1
            self.front -= 1
            # Get the first item of the queue
            return self.queue.pop()

    def peek(self):
        """
        Method to get the next snapshot in the queue without
        popping it from the stack.

        Returns None if the stack size is smaller than 1.
        """

        if len(self.queue) == 0:
            return None

        return self.queue[self.front]

    def empty(self):
        return len(self.queue) == 0

    def __str__(self):
        return str(self.queue)



from .order_executor import OrderExecutor
from .binance import BinanceOrderExecutor
from investing_algorithm_framework.core.order_executors.factory import \
    DefaultOrderExecutorFactory

__all__ = [
    "OrderExecutor", "BinanceOrderExecutor", "DefaultOrderExecutorFactory"
]



import logging

from flask import Blueprint, jsonify, request

from investing_algorithm_framework import current_app
from investing_algorithm_framework.configuration.constants \
    import TIME_FRAME_QUERY_PARAM
from investing_algorithm_framework.core.models import TimeFrame
from investing_algorithm_framework.schemas import PortfolioSerializer
from investing_algorithm_framework.views.utils import \
    get_query_param

logger = logging.getLogger(__name__)

blueprint = Blueprint("portfolio-views", __name__)


@blueprint.route("/api/portfolios/<string:identifier>", methods=["GET"])
def retrieve(identifier="default"):
    """
    View for retrieving of an portfolio of the algorithm.
    """

    # Get the default portfolio
    if identifier == "default":
        portfolio = current_app.algorithm\
            .get_portfolio_manager()\
            .get_portfolio()
    else:
        portfolio = current_app.algorithm \
            .get_portfolio_manager(identifier=identifier) \
            .get_portfolio()

    time_frame = get_query_param(
        TIME_FRAME_QUERY_PARAM, request.args, TimeFrame.ONE_DAY.value
    )

    return jsonify(
        PortfolioSerializer(context={"time_frame": time_frame})
        .dump(portfolio)
    ), 200



from abc import abstractmethod


class Position:

    @abstractmethod
    def get_symbol(self):
        pass

    @abstractmethod
    def get_orders(self):
        pass

    @abstractmethod
    def get_amount(self):
        pass

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def to_string(self):
        return self.repr(
            symbol=self.get_symbol(),
            amount=self.get_amount()
        )

    def __repr__(self):
        return self.to_string()


from investing_algorithm_framework.core.models import TimeFrame
from investing_algorithm_framework.core.models.snapshots \
    import SQLLitePortfolioSnapshot, \
    SQLLiteAssetPriceHistory
from investing_algorithm_framework.core.performance.asset_price_queue \
    import AssetPricesQueue
from investing_algorithm_framework.core.performance.intervals_queue \
    import IntervalsQueue
from investing_algorithm_framework.core.performance.snapshot_queue \
    import SnapShotQueue


class SnapshotAssetPriceCollection:
    """
    SnapshotAssetPriceCollection is a class that implements the iterator design
    pattern to match portfolio snapshots with asset prices based on
    specific time intervals.

    Iterating over this data structure, gives you a time interval, a snapshot
    and a set of asset prices
    """

    def __init__(self, portfolio, time_frame: TimeFrame):
        self.time_frame = TimeFrame.from_value(time_frame).value
        self.portfolio = portfolio

        # Retrieve snapshots
        self.snapshots = self._retrieve_snapshots(portfolio, time_frame)
        self.asset_prices = self._retrieve_asset_price_histories(
            self._retrieve_unique_symbols(), self.time_frame
        )
        self.current_snapshot = None
        self.current_date = None
        self.index = 0
        self.intervals = IntervalsQueue(self.time_frame)

    def __iter__(self):
        self.snapshot_queue = SnapShotQueue(self.snapshots)

        self.intervals = IntervalsQueue(self.time_frame)
        self.asset_prices_queue = AssetPricesQueue(
            self.asset_prices, self.time_frame
        )
        self.current_snapshot = self.snapshot_queue.pop()
        self.current_date, self.current_asset_prices = \
            self.asset_prices_queue.pop()
        self.first_snapshot = self.current_snapshot
        return self

    def __next__(self):

        if self.asset_prices_queue.empty():
            raise StopIteration

        peek_snapshot = self.snapshot_queue.peek()

        # Check if the current snapshot is the pre-first one
        if self.first_snapshot.id == self.current_snapshot.id:

            # There is no snapshot for this time interval
            if self.current_snapshot.created_at > self.current_date:
                datetime = self.current_date
                asset_prices = self.current_asset_prices
                self.current_date, self.current_asset_prices = \
                    self.asset_prices_queue.pop()
                return datetime, None, asset_prices

            if peek_snapshot is not None \
                    and peek_snapshot.created_at >= self.current_date:
                datetime = self.current_date
                asset_prices = self.current_asset_prices
                snapshot = self.current_snapshot
                self.current_date, self.current_asset_prices = \
                    self.asset_prices_queue.pop()
                self.current_snapshot = self.snapshot_queue.pop()
                return datetime, snapshot, asset_prices
            else:
                datetime = self.current_date
                asset_prices = self.current_asset_prices
                self.current_date, self.current_asset_prices = \
                    self.asset_prices_queue.pop()
                return datetime, self.current_snapshot, asset_prices

        # Check if we need to pop prices and interval time
        if peek_snapshot is not None:

            if peek_snapshot.created_at >= self.current_date:
                snapshot = self.current_snapshot
                interval = self.current_date
                asset_prices = self.current_asset_prices

                # Pop from queues
                self.current_snapshot = self.snapshot_queue.pop()
                self.current_date, self.asset_prices \
                    = self.asset_prices_queue.pop()

                return interval, snapshot, asset_prices
            else:
                snapshot = self.current_snapshot

                # Make known to the snapshot that it is an inner snapshot
                snapshot.set_inner_snapshot(True)

                # Pop snapshot from queue
                self.current_snapshot = self.snapshot_queue.pop()
                return snapshot.created_at, snapshot, self.asset_prices
        else:
            datetime = self.current_date
            asset_prices = self.current_asset_prices
            self.current_date, self.current_asset_prices = \
                self.asset_prices_queue.pop()
            return datetime, self.current_snapshot, asset_prices

    def _retrieve_unique_symbols(self):
        unique_symbols = []

        # Get unique symbols
        for portfolio_snapshot in self.snapshots:
            for position in portfolio_snapshot.positions:
                if position.symbol not in unique_symbols:
                    unique_symbols.append(position.symbol)

        return unique_symbols

    def _retrieve_asset_price_histories(self, symbols, time_frame):
        asset_prices = []

        # Retrieve and save the price histories
        for symbol in symbols:
            asset_price_history = SQLLiteAssetPriceHistory.of(
                market=self.portfolio.market,
                target_symbol=symbol,
                trading_symbol=self.portfolio.trading_symbol,
                time_frame=time_frame
            )

            asset_prices.append(asset_price_history.get_prices())
        return asset_prices

    def _retrieve_snapshots(self, portfolio, time_frame):

        start_datetime, end_datetime = time_frame.create_time_frame()

        snapshots = SQLLitePortfolioSnapshot.query\
            .filter(SQLLitePortfolioSnapshot.created_at >= start_datetime)\
            .filter(SQLLitePortfolioSnapshot.created_at <= end_datetime)\
            .order_by(SQLLitePortfolioSnapshot.created_at.desc())\
            .all()

        if(SQLLitePortfolioSnapshot.query
                .filter_by(portfolio_id=portfolio.id)
                .filter(SQLLitePortfolioSnapshot.created_at < start_datetime)
                .count() > 0):
            additional_snapshot = SQLLitePortfolioSnapshot.query \
                .filter_by(portfolio_id=portfolio.id)\
                .filter(SQLLitePortfolioSnapshot.created_at < start_datetime) \
                .order_by(SQLLitePortfolioSnapshot.created_at.asc())\
                .first()

            snapshots.append(additional_snapshot)

        return snapshots



from .order_validator import OrderValidator
from .default_order_validator import DefaultOrderValidator


class OrderValidatorFactory:

    @staticmethod
    def of(market) -> OrderValidator:
        return DefaultOrderValidator()



from investing_algorithm_framework.core.market_services import \
    BinanceMarketService
from investing_algorithm_framework.core.models.data_provider import Ticker, \
    OrderBook
from investing_algorithm_framework.configuration.constants import BINANCE


class BinanceDataProviderMixin(BinanceMarketService):
    identifier = BINANCE

    def provide_ticker(
            self, target_symbol, trading_symbol, algorithm_context, **kwargs
    ) -> Ticker:
        market_service = algorithm_context.get_market_service(BINANCE)
        data = market_service.get_ticker(target_symbol, trading_symbol)

        return Ticker(
            symbol=f"{target_symbol.upper()}{trading_symbol.upper()}",
            price=data["ask"],
            ask_price=data["ask"],
            ask_volume=data["askVolume"],
            bid_price=data["bid"],
            bid_volume=data["bidVolume"],
            high_price=data["high"],
            low_price=data["low"]
        )

    def provide_order_book(
            self, target_symbol, trading_symbol, algorithm_context, **kwargs
    ) -> OrderBook:
        market_service = algorithm_context.get_market_service(BINANCE)
        data = market_service.get_order_book(target_symbol, trading_symbol)
        return OrderBook(
            symbol=f"{target_symbol.upper()}{trading_symbol.upper()}",
            bids=data["bids"],
            asks=data["asks"]
        )



class ApiException(Exception):

    def __init__(self, message: str = None, status_code: int = 400):
        super(ApiException, self).__init__(message)
        self._message = message
        self._status_code = status_code

    @property
    def error_message(self):

        if self._message is None:
            return "An error occurred"

        return self._message

    @property
    def status_code(self):

        if self._status_code is None:
            return 500

        return self._status_code


class PermissionDeniedApiException(ApiException):

    def __init__(self):
        super(PermissionDeniedApiException, self).__init__(
            message="Permission denied", status_code=401
        )



from marshmallow import Schema, fields


class PositionSerializer(Schema):
    id = fields.Int(dump_only=True)
    symbol = fields.String(dump_only=True)
    amount = fields.Float(dump_only=True)

    # Optional fields
    identifier = fields.Method("get_identifier")
    orders = fields.Method("get_orders")

    @staticmethod
    def get_identifier(obj):
        return obj.portfolio.identifier

    @staticmethod
    def get_orders(obj):
        return obj.orders.count()



from enum import Enum
from datetime import datetime, timedelta
from dateutil import relativedelta


class TimeFrame(Enum):
    CURRENT = "CURRENT"
    ONE_HOUR = "ONE_HOUR"
    ONE_DAY = "ONE_DAY"
    ONE_WEEK = "ONE_WEEK"
    ONE_MONTH = "ONE_MONTH"
    ONE_YEAR = "ONE_YEAR"

    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):

            for entry in TimeFrame:

                if value.upper() == entry.value:
                    return entry

            raise ValueError(
                f"Could not convert {value} to TimeFrame"
            )

    @staticmethod
    def from_value(value):

        if isinstance(value, str):
            return TimeFrame.from_string(value)

        if isinstance(value, TimeFrame):

            for entry in TimeFrame:

                if value == entry:
                    return entry

        raise ValueError(
            f"Could not convert {value} to TimeFrame"
        )

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value
        else:
            return TimeFrame.from_string(other) == self

    def create_time_frame(self, end_datetime=None):

        if end_datetime is None:
            end_datetime = datetime.utcnow()

        if TimeFrame.ONE_HOUR.equals(self):
            start_datetime = end_datetime - \
                             relativedelta.relativedelta(hours=1)
        elif TimeFrame.ONE_DAY.equals(self):
            start_datetime = end_datetime - \
                             relativedelta.relativedelta(hours=24)
        elif TimeFrame.ONE_WEEK.equals(self):
            start_datetime = end_datetime - \
                             relativedelta.relativedelta(days=7)
        elif TimeFrame.ONE_MONTH.equals(self):
            start_datetime = \
                end_datetime - relativedelta.relativedelta(days=28)
        elif TimeFrame.ONE_YEAR.equals(self):
            start_datetime = \
                end_datetime - relativedelta.relativedelta(days=365)
        else:
            raise NotImplementedError(
                f"Timeframe {self.value} not implemented"
            )

        return start_datetime, end_datetime

    def duration(self):
        if TimeFrame.ONE_HOUR.equals(self):
            return relativedelta.relativedelta(hours=1)
        elif TimeFrame.ONE_DAY.equals(self):
            return relativedelta.relativedelta(hours=24)
        elif TimeFrame.ONE_WEEK.equals(self):
            return relativedelta.relativedelta(days=7)
        elif TimeFrame.ONE_MONTH.equals(self):
            return relativedelta.relativedelta(days=28)
        elif TimeFrame.ONE_YEAR.equals(self):
            return relativedelta.relativedelta(days=365)
        else:
            raise NotImplementedError(
                f"Timeframe {self.value} not implemented"
            )

    @property
    def time_interval(self):
        from investing_algorithm_framework.core.models import TimeInterval

        if TimeFrame.CURRENT.equals(self):
            return TimeInterval.CURRENT
        elif TimeFrame.ONE_HOUR.equals(self):
            return TimeInterval.MINUTES_ONE
        elif TimeFrame.ONE_DAY.equals(self):
            return TimeInterval.MINUTES_FIFTEEN
        elif TimeFrame.ONE_WEEK.equals(self):
            return TimeInterval.HOURS_ONE
        elif TimeFrame.ONE_MONTH.equals(self):
            return TimeInterval.HOURS_FOUR
        elif TimeFrame.ONE_YEAR.equals(self):
            return TimeInterval.DAYS_ONE
        else:
            raise NotImplementedError(
                f"Timeframe {self.value} not implemented"
            )

    @property
    def intervals(self):
        intervals = []
        start_datetime, end_datetime = self.create_time_frame()
        delta = end_datetime - start_datetime

        intervals.append(end_datetime)

        if TimeFrame.ONE_HOUR.equals(self):

            for i in range(1, int(delta.total_seconds() / 60)):
                intervals.append(end_datetime - timedelta(minutes=i))

        elif TimeFrame.ONE_DAY.equals(self):

            for i in range(1, int((delta.total_seconds() / 60) / 15)):
                intervals.append(end_datetime - timedelta(minutes=i * 15))

        elif TimeFrame.ONE_WEEK.equals(self):

            for i in range(1, int((delta.total_seconds() / 60) / 60)):
                intervals.append(end_datetime - timedelta(hours=i))

        elif TimeFrame.ONE_MONTH.equals(self):

            for i in range(1, int(((delta.total_seconds() / 60) / 60) / 4)):
                intervals.append(end_datetime - timedelta(hours=i * 4))

        elif TimeFrame.ONE_YEAR.equals(self):

            for i in range(1, int(delta.days)):
                intervals.append(
                    end_datetime - relativedelta.relativedelta(days=i)
                )

        else:
            raise NotImplementedError(
                f"Timeframe {self.value} not implemented"
            )

        return intervals



from investing_algorithm_framework.core.data_providers.data_provider \
    import DataProvider
from investing_algorithm_framework.core.data_providers.binance_data_provider \
    import BinanceDataProvider
from investing_algorithm_framework.core.data_providers.factory import \
    DefaultDataProviderFactory


__all__ = [
    "DataProvider",
    "BinanceDataProvider",
    "DefaultDataProviderFactory"
]



from flask import Flask
from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()


def create_all_tables():
    db.create_all()


def initialize_db(app: Flask):
    db.init_app(app)
    db.app = app

from investing_algorithm_framework.core.models.performance_metric import \
    PerformanceMetric
from investing_algorithm_framework.core.models.order_status import OrderStatus
from investing_algorithm_framework.core.models.order_type import OrderType
from investing_algorithm_framework.core.models.order_side import OrderSide
from investing_algorithm_framework.core.models.time_unit import TimeUnit
from investing_algorithm_framework.core.models.order import Order
from investing_algorithm_framework.core.models.portfolio import Portfolio
from investing_algorithm_framework.core.models.position import Position
from investing_algorithm_framework.core.models.time_frame import TimeFrame
from investing_algorithm_framework.core.models.time_intervals import \
    TimeInterval
from investing_algorithm_framework.core.models.sql_lite import \
    SQLLitePosition, SQLLiteOrder, SQLLitePortfolio
from investing_algorithm_framework.core.models.snapshots import \
    PositionSnapshot, PortfolioSnapshot, SQLLitePortfolioSnapshot, \
    SQLLiteAssetPrice, SQLLitePositionSnapshot, SQLLiteAssetPriceHistory, \
    AssetPrice, AssetPriceHistory
from investing_algorithm_framework.core.models.data_provider \
    import TradingDataTypes

__all__ = [
    "db",
    "Portfolio",
    "Position",
    "SQLLitePosition",
    'Order',
    "SQLLiteOrder",
    "OrderType",
    'OrderSide',
    "TimeUnit",
    "create_all_tables",
    "initialize_db",
    "OrderStatus",
    "TimeFrame",
    "TimeInterval",
    "PortfolioSnapshot",
    "PositionSnapshot",
    "PerformanceMetric",
    "SQLLitePortfolio",
    "SQLLitePortfolioSnapshot",
    "SQLLitePositionSnapshot",
    "AssetPrice",
    "AssetPriceHistory",
    "SQLLiteAssetPrice",
    "SQLLiteAssetPriceHistory",
    "TradingDataTypes"
]



from datetime import datetime
from random import randint

from sqlalchemy import UniqueConstraint, event
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import validates

from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import db, OrderSide, \
    OrderStatus, OrderType, Portfolio
from investing_algorithm_framework.core.models.model_extension \
    import SQLAlchemyModelExtension
from investing_algorithm_framework.core.models.sql_lite import SQLLiteOrder
from investing_algorithm_framework.core.order_validators import \
    OrderValidatorFactory


def random_id():
    """
    Function to create a random ID. This function checks first if
    the generated ID is not already taken.
    Returns: random integer that can be used as an ID
    """
    minimal = 100
    maximal = 1000000000000000000
    rand = randint(minimal, maximal)

    while SQLLitePortfolio.query.filter_by(id=rand).first() is not None:
        rand = randint(minimal, maximal)

    return rand


class SQLLitePortfolio(db.Model, Portfolio, SQLAlchemyModelExtension):

    """
    Class Portfolio: a database model for an
    AlgorithmPortfolio instance.

    Attributes:
    A AlgorithmPortfolio instance consists out of the following attributes:

    - id: unique identification
    - algorithm_id: reference to the Algorithm instance (Algorithm Service)
    - free_fraction: fraction of the AlgorithmPortfolio that can be
    freely invested
    - invested_fraction: fraction of the AlgorithmPortfolio that is
    already invested
    - net_profit_percentage: lifetime return of the AlgorithmPortfolio
    - created_at: The datetime the portfolio was created
    - updated_at: The datetime the portfolio was updated

    Relationships:
    - algorithm_orders: all the AlgorithmOrders which belong to the
    AlgorithmPortfolio
    - algorithm_positions: all the AlgorithmPositions which belong to the
    AlgorithmPortfolio

    During creation of a AlgorithmPortfolio, you need to provide an
    Algorithm Id.
    """

    __tablename__ = "portfolios"

    id = db.Column(db.Integer, primary_key=True)
    market = db.Column(db.String, nullable=False)
    identifier = db.Column(db.String, nullable=False)
    trading_symbol = db.Column(db.String, nullable=False)
    unallocated = db.Column(db.Float, nullable=False, default=0)
    realized = db.Column(db.Float, nullable=False, default=0)

    total_revenue = db.Column(db.Float, nullable=False, default=0)
    total_cost = db.Column(db.Float, nullable=False, default=0)

    created_at = db.Column(db.DateTime, default=datetime.utcnow())
    updated_at = db.Column(
        db.DateTime,
        default=datetime.utcnow(),
        onupdate=datetime.utcnow()
    )

    # Relationships
    positions = db.relationship(
        "SQLLitePosition",
        back_populates="portfolio",
        lazy="dynamic",
        cascade="all,delete",
    )

    # Constraints
    __table_args__ = (
        UniqueConstraint(
            'trading_symbol',
            'identifier',
            name='_trading_currency_identifier_uc'
        ),
    )

    @validates('trading_symbol')
    def _write_once(self, key, value):
        existing = getattr(self, key)

        if existing is not None:
            raise ValueError("{} is write-once".format(key))
        return value

    def __init__(
            self, trading_symbol, unallocated, identifier, market, **kwargs
    ):
        self.id = random_id()
        self.identifier = identifier
        self.trading_symbol = trading_symbol
        self.unallocated = unallocated
        self.realized = 0
        self.total_revenue = 0
        self.total_cost = 0
        self.market = market
        self.created_at = datetime.utcnow()
        super(Portfolio, self).__init__(**kwargs)

    @hybrid_property
    def allocated(self):
        from investing_algorithm_framework.core.models import SQLLitePosition

        position_ids = self.positions.with_entities(SQLLitePosition.id)

        orders = SQLLiteOrder.query \
            .filter_by(order_side=OrderSide.BUY.value) \
            .filter_by(status=OrderStatus.SUCCESS.value)\
            .filter(SQLLiteOrder.position_id.in_(position_ids)) \
            .all()

        allocated = 0

        for order in orders:
            allocated += order.current_value

        return allocated

    @hybrid_property
    def allocated_percentage(self):

        # Prevent zero division
        if self.allocated == 0 and self.unallocated == 0:
            return 0

        return (self.allocated / (self.allocated + self.unallocated)) * 100

    @hybrid_property
    def unallocated_percentage(self):

        # Prevent zero division
        if self.allocated == 0 and self.unallocated == 0:
            return 0

        return (self.unallocated / self.unallocated + self.allocated) * 100

    def add_order(self, order):
        self._validate_order(order)

        self._add_order_to_position(order)

        order.update(db, {"status": OrderStatus.TO_BE_SENT.value}, commit=True)

        if OrderSide.BUY.equals(order.order_side):
            self.unallocated -= order.amount_trading_symbol

            # Create the snapshot at the time the order was created
            self.snapshot(creation_datetime=order.created_at)

        db.session.commit()

    def _validate_order(self, order):
        order_validator = OrderValidatorFactory.of(self.market)
        order_validator.validate(order, self)

    def _add_order_to_position(self, order):
        from investing_algorithm_framework.core.models import SQLLitePosition

        position = SQLLitePosition.query \
            .filter_by(symbol=order.target_symbol) \
            .filter_by(portfolio=self) \
            .first()

        if OrderSide.BUY.equals(order.order_side):

            if position is None:
                position = SQLLitePosition(symbol=order.target_symbol)
                position.save(db)
                self.positions.append(position)
                db.session.commit()
        else:

            if position is None:
                raise OperationalException(
                    "Sell order can 't be added to non existing position"
                )

        position.orders.append(order)
        db.session.commit()

    def create_order(
        self,
        context,
        order_type,
        symbol,
        price=None,
        amount_trading_symbol=None,
        amount_target_symbol=None,
        order_side=OrderSide.BUY.value,
    ):
        if OrderType.MARKET.equals(order_type):

            if OrderSide.SELL.equals(order_side):
                order = SQLLiteOrder(
                    target_symbol=symbol,
                    trading_symbol=self.trading_symbol,
                    amount_target_symbol=amount_target_symbol,
                    order_type=OrderType.MARKET.value,
                    order_side=OrderSide.SELL.value,
                    price=price
                )
            else:
                raise OperationalException("Buy market order is not supported")
        else:
            order = SQLLiteOrder(
                target_symbol=symbol,
                trading_symbol=self.trading_symbol,
                amount_target_symbol=amount_target_symbol,
                amount_trading_symbol=amount_trading_symbol,
                order_type=order_type,
                order_side=order_side,
                price=price
            )

        return order

    def withdraw(self, amount, creation_datetime=datetime.utcnow()):

        if amount > self.unallocated:
            raise OperationalException(
                "Withdrawal is larger then unallocated size"
            )

        self.unallocated -= amount
        db.session.commit()
        self.snapshot(creation_datetime=creation_datetime, withdrawel=amount)

    def deposit(self, amount, creation_datetime=datetime.utcnow()):
        self.unallocated += amount
        db.session.commit()
        self.snapshot(creation_datetime=creation_datetime, deposit=amount)

    def update(self, db, data, commit: bool = False, **kwargs):
        self.updated_at = datetime.utcnow()
        unallocated = data.pop("unallocated", None)

        if unallocated is not None:
            self.unallocated += unallocated
            self.snapshot()

        super(Portfolio, self).update(db, data, **kwargs)

    def snapshot(
        self, withdrawel=0, deposit=0, commit=True, creation_datetime=None
    ):
        from investing_algorithm_framework.core.models.snapshots \
            import SQLLitePortfolioSnapshot

        SQLLitePortfolioSnapshot\
            .from_portfolio(
                self, creation_datetime, withdrawel=withdrawel, deposit=deposit
            ).save(db, commit=commit)

    def get_id(self):
        return self.id

    def get_identifier(self):
        return self.identifier

    def get_trading_symbol(self):
        return self.trading_symbol

    def get_positions(self, symbol: str = None):
        query_set = self.positions

        if symbol is not None:
            query_set = query_set.filter_by(symbol=symbol)

        return query_set.all()

    def get_number_of_positions(self):
        return self.positions.count()

    def get_number_of_orders(self):
        from investing_algorithm_framework import SQLLitePosition

        position_ids = self.positions.with_entities(SQLLitePosition.id)

        return SQLLiteOrder.query \
            .filter(SQLLiteOrder.position_id.in_(position_ids)) \
            .count()

    def get_orders(
        self,
        status: OrderStatus = None,
        side: OrderSide = None,
        target_symbol: str = None,
        trading_symbol: str = None,
        lazy: bool = False
    ):
        from investing_algorithm_framework.core.models import SQLLitePosition

        position_ids = self.positions.with_entities(SQLLitePosition.id)

        query_set = SQLLiteOrder.query \
            .filter(SQLLiteOrder.position_id.in_(position_ids))

        if status:
            status = OrderStatus.from_value(status)
            query_set = query_set.filter_by(status=status.value)

        if target_symbol:
            query_set = query_set.filter_by(target_symbol=target_symbol)

        if trading_symbol:
            query_set = query_set.filter_by(trading_symbol=trading_symbol)

        if side:
            query_set = query_set.filter_by(order_side=side)

        if lazy:
            return query_set

        return query_set.all()

    def get_unallocated(self):
        return self.unallocated

    def get_allocated(self):
        return self.allocated

    def get_realized(self):
        return self.realized

    def get_total_revenue(self):
        return self.total_revenue

    @staticmethod
    def from_dict(data):
        orders = data.get("orders")
        positions = data.get("positions")

    def to_dict(self):
        orders = self.get_orders()
        positions = self.get_positions()

    def __repr__(self):
        return self.to_string()


# first snapshot on creation of a portfolio
@event.listens_for(SQLLitePortfolio, 'before_insert')
def snapshot_after_insert(mapper, connection, portfolio):
    portfolio.snapshot(commit=False)



from abc import ABC, abstractmethod
from typing import List

from investing_algorithm_framework.core.events.observer import Observer


class Observable(ABC):
    """
    Class Observable: manages and updates it's observers.
    """

    def __init__(self) -> None:
        self._observers: List[Observer] = []

    @abstractmethod
    def add_observer(self, observer: Observer) -> None:

        if isinstance(observer, Observer) and observer not in self._observers:
            self._observers.append(observer)

    @abstractmethod
    def remove_observer(self, observer: Observer) -> None:

        if observer in self._observers:
            self._observers.remove(observer)

    def notify_observers(self, **kwargs) -> None:

        for observer in self._observers:
            observer.update(self, **kwargs)

    @property
    def observers(self) -> List[Observer]:
        return self._observers



from investing_algorithm_framework.core.models import TimeFrame


class IntervalsQueue:
    """
    Time intervals queue for logical order of time intervals
    """

    def __init__(self, time_frame):
        self.intervals = TimeFrame.from_value(time_frame).intervals
        self.size = len(self.intervals)
        self.front = self.size - 1

    def pop(self):

        if len(self.intervals) == 0:
            return None
        else:
            interval = self.intervals.pop()
            return interval

    def peek(self):
        """
        Method to get the next time interval in the queue without
        popping it from the stack.

        Returns None if the stack size is smaller than 1.
        """

        if len(self.intervals) == 0:
            return None, None

        return self.intervals[self.front]

    def empty(self):
        return len(self.intervals) == 0

    def __str__(self):
        return str(self.intervals)



class ImproperlyConfigured(Exception):
    """
    Class ImproperlyConfigured: Exception class indicating a problem with
    the configuration of the investing_algorithm_framework
    """
    def __init__(self, message) -> None:
        super(ImproperlyConfigured, self).__init__(message)
        self.error_message = message


class OperationalException(Exception):
    """
    Class OperationalException: Exception class indicating a problem occurred
    during running of the investing_algorithm_framework
    """
    def __init__(self, message) -> None:
        super(OperationalException, self).__init__(message)
        self.error_message = message



from flask import jsonify


def normalize_query_param(value):
    """
    Given a non-flattened query parameter value,
    and if the value is a list only containing 1 item,
    then the value is flattened.

    :param value: a value from a query parameter
    :return: a normalized query parameter value
    """

    if len(value) == 1 and value[0].lower() in ["true", "false"]:

        if value[0].lower() == "true":
            return True
        return False

    return value if len(value) > 1 else value[0]


def normalize_query(params):
    """
    Converts query parameters from only containing one value for each parameter,
    to include parameters with multiple values as lists.

    :param params: a flask query parameters data_provider structure
    :return: a dict of normalized query parameters
    """
    params_non_flat = params.to_dict(flat=False)
    return {k: normalize_query_param(v) for k, v in params_non_flat.items()}


def has_query_param(key, params):
    query_params = normalize_query(params)
    return key in query_params


def get_query_param(key, params, default=None, many=False):
    query_params = normalize_query(params)
    selection = query_params.get(key, default)

    if isinstance(selection, list) and len(selection) > 1 and not many:
        return selection[0]

    return selection


def split(list_a, chunk_size):

    for i in range(0, len(list_a), chunk_size):
        yield list_a[i:i + chunk_size]


def create_paginated_response(query_set, serializer, page=1, per_page=20):
    """
    Creates a paginated response from a query set. The given query set
    is paginated the function.

    :param query_set: a lazy query
    :param serializer: an instance of an marshmallow schema
    :return: a json of paginated query
    """

    if isinstance(query_set, list):
        chunks = list(split(query_set, per_page))

        if len(chunks) > 0:
            items = chunks[page - 1]
        else:
            items = []

        return jsonify(
            {
                'total': len(query_set),
                'page': page,
                'per_page': per_page,
                'items': serializer.dump(items, many=True)
            }
        )
    else:
        paginated_query_set = query_set.paginate()

    return jsonify(
        {
            'total': paginated_query_set.total,
            'page': paginated_query_set.page,
            'per_page': paginated_query_set.per_page,
            'items': serializer.dump(paginated_query_set.items, many=True)
        }
    )



from abc import abstractmethod, ABC

from investing_algorithm_framework.core.models import Order, OrderStatus
from investing_algorithm_framework.core.identifier import Identifier


class OrderExecutor(ABC, Identifier):

    def initialize(self, algorithm_context, throw_exception=True):
        pass

    @abstractmethod
    def execute_limit_order(
        self, order: Order, algorithm_context, **kwargs
    ) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def execute_market_order(
        self, order: Order, algorithm_context, **kwargs
    ) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get_order_status(
        self, order: Order, algorithm_context, **kwargs
    ) -> OrderStatus:
        raise NotImplementedError()



from datetime import datetime

from investing_algorithm_framework.core.models.model_extension import \
    SQLAlchemyModelExtension


class Ticker(SQLAlchemyModelExtension):

    def __init__(
            self,
            symbol,
            price,
            ask_price=None,
            ask_volume=None,
            bid_price=None,
            bid_volume=None,
            high_price=None,
            low_price=None,
            creation_date=None
    ):
        self.symbol = symbol
        self.price = price
        self.ask_price = ask_price
        self.ask_volume = ask_volume
        self.bid_price = bid_price
        self.bid_volume = bid_volume
        self.high_price = high_price
        self.low_price = low_price

        if creation_date is None:
            self.creation_date = datetime.now()

    def get_price(self):
        return self.price

    def get_creation_date(self):
        return self.creation_date

    def get_symbol(self):
        return self.symbol

    def to_dict(self):
        return {
            "symbol": self.symbol,
            "price": self.price,
            "ask_price": self.ask_price,
            "ask_volume": self.ask_volume,
            "bid_price": self.bid_price,
            "bid_volume": self.bid_volume,
            "high_price": self.high_price,
            "low_price": self.low_price
        }

    @staticmethod
    def from_dict(data):
        return Ticker(
            symbol=data.get("symbol"),
            price=data.get("price"),
            ask_price=data.get("ask_price"),
            ask_volume=data.get("ask_volume"),
            bid_price=data.get("bid_price"),
            bid_volume=data.get("bid_volume"),
            high_price=data.get("high_price"),
            low_price=data.get("low_price"),
            creation_date=data.get("creation_date")
        )

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def __repr__(self):
        return self.repr(
            symbol=self.symbol,
            price=self.price,
            ask_price=self.ask_price,
            ask_volume=self.ask_volume,
            bid_price=self.bid_price,
            bid_volume=self.bid_volume,
            high_price=self.high_price,
            low_price=self.low_price
        )



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.data_providers.binance_data_provider \
    import BinanceDataProvider


class DefaultDataProviderFactory:

    @staticmethod
    def of_identifier(identifier):
        if identifier.upper() == BINANCE:
            return BinanceDataProvider()

        return None



from abc import ABC, abstractmethod


class Observer(ABC):
    """
    Class Observer: Receive updates from it's observables.
    """

    @abstractmethod
    def update(self, observable, **kwargs) -> None:
        pass



import os
from investing_algorithm_framework.configuration.settings import Config
from investing_algorithm_framework.configuration.validator import \
    ConfigValidator
from investing_algorithm_framework.configuration.setup import create_app, \
    setup_config, setup_database, setup_logging
from investing_algorithm_framework.configuration.constants import \
    RESOURCES_DIRECTORY, LOG_LEVEL

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


__all__ = [
    "Config",
    "BASE_DIR",
    "ConfigValidator",
    "create_app",
    "setup_database",
    "setup_config",
    "RESOURCES_DIRECTORY",
    "LOG_LEVEL",
    "setup_logging"
]



from investing_algorithm_framework.core.models.snapshots.position_snapshot \
    import PositionSnapshot, SQLLitePositionSnapshot
from investing_algorithm_framework.core.models.snapshots.portfolio_snapshot \
    import PortfolioSnapshot, SQLLitePortfolioSnapshot
from investing_algorithm_framework.core.models.snapshots.asset_price import \
    AssetPrice, SQLLiteAssetPrice
from investing_algorithm_framework.core.models.snapshots.asset_price_history \
    import AssetPriceHistory, SQLLiteAssetPriceHistory


__all__ = [
    "SQLLiteAssetPrice",
    "SQLLiteAssetPriceHistory",
    "PositionSnapshot",
    "PortfolioSnapshot",
    "AssetPrice",
    "AssetPriceHistory",
    "SQLLitePortfolioSnapshot",
    "SQLLitePositionSnapshot"
]



from investing_algorithm_framework.core.models.sql_lite.order import \
    SQLLiteOrder
from investing_algorithm_framework.core.models.sql_lite.position import \
    SQLLitePosition
from investing_algorithm_framework.core.models.sql_lite.portfolio import \
    SQLLitePortfolio

__all__ = [
    "SQLLiteOrder", "SQLLitePosition", "SQLLitePortfolio"
]


import logging

import sqlalchemy
from flask import Blueprint, request

from investing_algorithm_framework.configuration.constants import\
    SYMBOL_QUERY_PARAM, IDENTIFIER_QUERY_PARAM
from investing_algorithm_framework import Position, Portfolio, db, \
    current_app, OperationalException, ApiException
from investing_algorithm_framework.schemas import PositionSerializer
from investing_algorithm_framework.views.utils import normalize_query, \
    create_paginated_response, get_query_param

logger = logging.getLogger(__name__)

blueprint = Blueprint("position-views", __name__)


def apply_position_query_parameters(query_set):
    query_params = normalize_query(request.args)

    if SYMBOL_QUERY_PARAM in query_params:
        query_set = query_set.filter_by(
            symbol=query_params[SYMBOL_QUERY_PARAM]
        )

    if IDENTIFIER_QUERY_PARAM in query_params:
        portfolio = Portfolio.query\
            .filter_by(identifier=query_params[IDENTIFIER_QUERY_PARAM])\
            .first()

        if portfolio is None:
            return db.session.query(Position).filter(sqlalchemy.sql.false())

        query_set = query_set.filter_by(portfolio=portfolio)

    return query_set


@blueprint.route("/api/positions", methods=["GET"])
def list_positions():
    """
    View for listing of the positions of the algorithm. This view will list all
    the positions of all portfolios of your algorithm.

    You can provide to this view the following query params:
        - symbol: the symbol that is traded in the orders that belong to the
        position.
    The response in the view is paginated.
    """
    try:

        identifier = get_query_param(IDENTIFIER_QUERY_PARAM, request.args, None)

        # Get the default portfolio
        if identifier == "default":
            portfolio = current_app.algorithm \
                .get_portfolio_manager() \
                .get_portfolio()
        else:
            portfolio = current_app.algorithm \
                .get_portfolio_manager(identifier=identifier) \
                .get_portfolio()
    except OperationalException as e:
        raise ApiException(e.error_message, status_code=404)

    symbol = get_query_param(SYMBOL_QUERY_PARAM, request.args, None)

    positions = portfolio.get_positions(symbol)

    # Create serializer
    serializer = PositionSerializer()

    # Paginate query
    return create_paginated_response(positions, serializer), 200



from abc import abstractmethod, ABC

from investing_algorithm_framework.configuration.constants import \
    TRADING_SYMBOL
from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.identifier import Identifier
from investing_algorithm_framework.core.market_identifier import \
    MarketIdentifier
from investing_algorithm_framework.core.models import OrderSide, OrderType, \
    Portfolio


class PortfolioManager(ABC, Identifier, MarketIdentifier):
    trading_symbol = None

    @abstractmethod
    def get_unallocated(self, algorithm_context, sync=False):
        pass

    @abstractmethod
    def get_allocated(self, algorithm_context, sync=False):
        pass

    @abstractmethod
    def initialize(self, algorithm_context):
        pass

    @abstractmethod
    def get_portfolio(self, algorithm_context) -> Portfolio:
        pass

    def get_trading_symbol(self, algorithm_context) -> str:
        trading_symbol = getattr(self, "trading_symbol", None)

        if trading_symbol is None:
            trading_symbol = algorithm_context.config.get(TRADING_SYMBOL, None)

        if trading_symbol is None:
            raise OperationalException(
                "Trading symbol is not set. Either override "
                "'get_trading_symbol' method or set "
                "the 'trading_symbol' attribute in the algorithm config."
            )

        return trading_symbol

    @abstractmethod
    def get_positions(self, symbol: str = None, lazy=False):
        pass

    @abstractmethod
    def get_orders(self, symbol: str = None, status=None, lazy=False):
        pass

    @abstractmethod
    def create_order(
        self,
        symbol,
        price=None,
        amount_trading_symbol=None,
        amount_target_symbol=None,
        order_type=OrderType.LIMIT.value,
        order_side=OrderSide.BUY.value,
        context=None,
        validate_pair=True,
    ):
        pass

    @abstractmethod
    def add_order(self, order):
        pass



from investing_algorithm_framework.schemas.order import OrderSerializer
from investing_algorithm_framework.schemas.position import PositionSerializer
from investing_algorithm_framework.schemas.portfolio import PortfolioSerializer

__all__ = [
    "OrderSerializer",
    "PositionSerializer",
    "PortfolioSerializer"
]



from abc import abstractmethod
from sqlalchemy import UniqueConstraint, or_
from datetime import datetime
from investing_algorithm_framework.core.models.model_extension import \
    SQLAlchemyModelExtension
from investing_algorithm_framework.core.models import db, \
    SQLLiteOrder, OrderStatus, \
    OrderSide
from investing_algorithm_framework.core.models.snapshots import \
    PositionSnapshot


class PortfolioSnapshot:

    @abstractmethod
    def get_portfolio_id(self):
        pass

    @abstractmethod
    def get_market(self):
        pass

    @abstractmethod
    def get_realized(self):
        pass

    @abstractmethod
    def get_pending_value(self):
        pass

    @abstractmethod
    def get_unallocated(self):
        pass

    @abstractmethod
    def get_allocated(self, asset_prices=None):
        pass

    @abstractmethod
    def get_total_revenue(self):
        pass

    @abstractmethod
    def get_created_at(self):
        pass

    @abstractmethod
    def get_withdrawel(self):
        pass

    @abstractmethod
    def get_deposit(self):
        pass

    @abstractmethod
    def set_inner_snapshot(self, flas):
        pass

    @abstractmethod
    def is_inner_snapshot(self):
        pass

    @abstractmethod
    def get_positions(self):
        pass

    @staticmethod
    @abstractmethod
    def from_portfolio(portfolio, creation_datetime=None, withdrawel=0, deposit=0):
        pass

    def get_total_value(self, asset_prices):
        if self.is_inner_snapshot():
            return self.get_pending_value() + self.get_unallocated() + self.get_allocated()

        total_value = 0

        total_value += self.get_allocated(asset_prices)
        total_value += self.get_pending_value()
        total_value += self.get_unallocated()
        return total_value

    @property
    def cashflow(self):

        if self.get_deposit() != 0:
            return self.get_deposit()

        if self.get_withdrawel() != 0:
            return -self.get_withdrawel()

        return 0

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def __repr__(self):
        return self.repr(
            portfolio_id=self.get_portfolio_id(),
            created_at=self.get_created_at(),
            market=self.get_market(),
        )


class SQLLitePortfolioSnapshot(PortfolioSnapshot, db.Model, SQLAlchemyModelExtension):

    __tablename__ = "portfolio_snapshots"

    id = db.Column(db.Integer, primary_key=True)
    portfolio_id = db.Column(db.Integer, nullable=False)
    market = db.Column(db.String, nullable=False)
    trading_symbol = db.Column(db.String, nullable=True)
    realized = db.Column(db.Float, default=0)
    allocated = db.Column(db.Float, default=0)
    unallocated = db.Column(db.Float, default=0)
    total_revenue = db.Column(db.Float, default=0)
    total_value = db.Column(db.Float, default=0)
    created_at = db.Column(db.DateTime, default=0)
    pending_value = db.Column(db.Float, default=0)

    # Cash flow attributes
    withdrawel = db.Column(db.Float, default=0)
    deposit = db.Column(db.Float, default=0)
    _inner_snapshot = db.Column(db.Boolean, default=False)

    # Relationships
    positions = db.relationship(
        "SQLLitePositionSnapshot",
        back_populates="portfolio",
        lazy="dynamic",
        cascade="all,delete",
    )

    __table_args__ = (
        UniqueConstraint(
            "portfolio_id",
            "created_at",
            name="_portfolio_snapshot_id_created_at_uc"
        ),
    )

    @staticmethod
    def from_portfolio(
        portfolio, creation_datetime=None, withdrawel=0, deposit=0
    ):
        snapshot = SQLLitePortfolioSnapshot()

        from investing_algorithm_framework.core.models import SQLLitePosition, \
            SQLLitePositionSnapshot

        positions = SQLLitePosition.query\
            .filter_by(portfolio_id=portfolio.id).all()

        position_snapshots = [
            SQLLitePositionSnapshot.from_position(position)
            for position in positions
        ]

        snapshot.total_revenue = portfolio.total_revenue
        snapshot.total_cost = portfolio.total_cost
        snapshot.positions = position_snapshots
        snapshot.portfolio_id = portfolio.id
        snapshot.trading_symbol = portfolio.trading_symbol
        snapshot.allocated = portfolio.allocated
        snapshot.realized = portfolio.realized
        snapshot.unallocated = portfolio.unallocated
        snapshot.created_at = creation_datetime
        snapshot.portfolio_id = portfolio.id
        snapshot.market = portfolio.market
        snapshot.withdrawel = withdrawel
        snapshot.deposit = deposit

        if snapshot.created_at is None:
            snapshot.created_at = datetime.utcnow()

        pending_orders = SQLLiteOrder.query\
            .filter(
                or_(
                    SQLLiteOrder.status == OrderStatus.PENDING.value,
                    SQLLiteOrder.status == OrderStatus.TO_BE_SENT.value
                )
            )\
            .filter_by(order_side=OrderSide.BUY.value)\
            .all()

        pending_value = 0

        for order in pending_orders:
            pending_value += order.amount_trading_symbol

        snapshot.pending_value = pending_value

        total_value = 0
        total_value += portfolio.allocated
        total_value += snapshot.pending_value
        total_value += snapshot.unallocated
        snapshot.total_value = total_value
        return snapshot

    @property
    def cash_flow(self):

        if self.deposit != 0:
            return self.deposit

        if self.withdrawel != 0:
            return -self.withdrawel

        return 0

    def is_inner_snapshot(self):
        return self._inner_snapshot

    def set_inner_snapshot(self, flag):
        self._inner_snapshot = flag

    def get_portfolio_id(self):
        return self.portfolio_id

    def get_market(self):
        return self.market

    def get_realized(self):
        return self.realized

    def get_pending_value(self):
        return self.pending_value

    def get_unallocated(self):
        return self.unallocated

    def get_allocated(self, asset_prices=None):

        if asset_prices is not None:
            allocated = 0

            for asset_price in asset_prices:
                position = self.positions \
                    .filter_by(symbol=asset_price.target_symbol).first()

                if position is not None:
                    allocated += position.amount * asset_price.price

            return allocated

        return self.allocated

    def get_total_revenue(self):
        return self.total_revenue

    def get_created_at(self):
        return self.created_at

    def get_withdrawel(self):
        return self.withdrawel

    def get_deposit(self):
        return self.deposit

    def get_positions(self):
        return self.positions



from .binance_api_secret_key_specifier import BinanceApiSecretKeySpecifierMixin
from .order_executors import BinanceOrderExecutorMixin
from .portfolio_managers import BinancePortfolioManagerMixin
from .data_providers import BinanceDataProviderMixin

__all__ = [
    "BinanceOrderExecutorMixin",
    "BinancePortfolioManagerMixin",
    "BinanceApiSecretKeySpecifierMixin",
    "BinanceDataProviderMixin"
]



from investing_algorithm_framework.core.performance.asset_price_queue \
    import AssetPricesQueue
from investing_algorithm_framework.core.performance.snapshot_queue \
    import SnapShotQueue
from investing_algorithm_framework.core.performance\
    .snapshot_asset_price_collection import SnapshotAssetPriceCollection
from investing_algorithm_framework.core.performance.performance_service \
    import PerformanceService
from investing_algorithm_framework.core.performance.intervals_queue \
    import IntervalsQueue


__all__ = [
    "AssetPricesQueue",
    "SnapShotQueue",
    "SnapshotAssetPriceCollection",
    "PerformanceService",
    "IntervalsQueue"
]


import logging
from abc import abstractmethod

from investing_algorithm_framework.core.models import OrderStatus

logger = logging.getLogger(__name__)


class Order:

    @abstractmethod
    def get_id(self):
        pass

    @abstractmethod
    def get_order_reference(self):
        pass

    @abstractmethod
    def set_order_reference(self, order_reference):
        pass

    @abstractmethod
    def get_target_symbol(self):
        pass

    @abstractmethod
    def get_trading_symbol(self):
        pass

    @abstractmethod
    def get_initial_price(self):
        pass

    @abstractmethod
    def set_initial_price(self, price):
        pass

    @abstractmethod
    def get_price(self):
        pass

    @abstractmethod
    def get_closing_price(self):
        pass

    @abstractmethod
    def set_closing_price(self, price):
        pass

    @abstractmethod
    def get_side(self):
        pass

    @abstractmethod
    def get_status(self) -> OrderStatus:
        pass

    @abstractmethod
    def set_status(self, status: OrderStatus):
        pass

    @abstractmethod
    def get_type(self):
        pass

    @abstractmethod
    def get_amount_target_symbol(self):
        pass

    @abstractmethod
    def get_amount_trading_symbol(self):
        pass

    @staticmethod
    @abstractmethod
    def from_dict(data):
        pass

    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def split(self, amount):
        pass

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def to_string(self):
        return self.repr(
            id=self.get_id(),
            order_reference=self.get_order_reference(),
            status=self.get_status(),
            initial_price=self.get_initial_price(),
            price=self.get_price(),
            closing_price=self.get_closing_price(),
            side=self.get_side(),
            type=self.get_type(),
            amount_target_symbol=self.get_amount_target_symbol(),
            amount_trading_symbol=self.get_amount_trading_symbol()
        )

    def __repr__(self):
        return self.to_string()



import logging
from investing_algorithm_framework.app import App
from investing_algorithm_framework.core.models import TimeUnit
from investing_algorithm_framework.configuration.constants import \
    CHECK_PENDING_ORDERS

logger = logging.getLogger("investing-algorithm-framework")

current_app: App = App()


# Check pending orders every 5 minutes
@current_app.algorithm.schedule(
    worker_id="default_order_checker", time_unit=TimeUnit.MINUTE, interval=5
)
def check_pending_orders(context):

    if context.config.get(CHECK_PENDING_ORDERS, False):
        logger.info("Checking pending orders")
        context.check_pending_orders()



from enum import Enum


class PerformanceMetric(Enum):
    OVERALL_PERFORMANCE = "OVERALL_PERFORMANCE"
    DELTA = "DELTA"

    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):

            for entry in PerformanceMetric:

                if value.upper() == entry.value:
                    return entry

            raise ValueError(
                f"Could not convert {value} to PerformanceMetric"
            )

    @staticmethod
    def from_value(value):

        if isinstance(value, str):
            return PerformanceMetric.from_string(value)

        if isinstance(value, PerformanceMetric):

            for entry in PerformanceMetric:

                if value == entry:
                    return entry

        raise ValueError(
            f"Could not convert {value} to PerformanceMetric"
        )

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value
        else:
            return PerformanceMetric.from_string(other) == self



from importlib import import_module
from typing import Any

from investing_algorithm_framework.core.exceptions \
    import ImproperlyConfigured, OperationalException


class AlgorithmContextConfiguration:
    """
    Base wrapper for ContextConfiguration module. It will load all the
    default settings for a given settings module and will allow for run time
    specification
    """

    def __init__(self) -> None:
        self.settings_module = None

    def load_settings_module(self, settings_module) -> None:
        self.settings_module = settings_module

        if self.settings_module is None:
            raise ImproperlyConfigured("There is no settings module defined")

        # Load the settings module
        module = import_module(self.settings_module)

        # Set all the attributes of the settings wrapper
        for setting in dir(module):

            if setting.isupper():
                setting_value = getattr(module, setting)
                setattr(self, setting, setting_value)

    def __getitem__(self, item) -> Any:

        if isinstance(item, str):

            if not hasattr(self, item):
                raise OperationalException(
                    "ContextConfig object doesn't have the specific "
                    "attribute {}".format(item)
                )

            return self.__getattribute__(item)
        else:
            raise OperationalException(
                "ContextConfig attributes can only be referenced by string"
            )

    def get(self, key: str, default: Any = None) -> Any:
        """
        Mimics the dict get() functionality
        """

        try:
            return self.__getitem__(key)
        # Ignore exception
        except OperationalException:
            pass

        return default

    def set(self, key: str, value: Any) -> None:

        if hasattr(self, key):
            raise OperationalException(
                "ContextConfig object already have the specific "
                "attribute {} specified".format(key)
            )

        setattr(self, key, value)



from investing_algorithm_framework.utils.version import get_version
from investing_algorithm_framework.core.context import \
    AlgorithmContextInitializer, AlgorithmContext
from investing_algorithm_framework.core.data_providers import DataProvider
from investing_algorithm_framework.core.portfolio_managers import \
    PortfolioManager, BinancePortfolioManager, SQLLitePortfolioManager
from investing_algorithm_framework.core.order_executors import OrderExecutor, \
    BinanceOrderExecutor
from investing_algorithm_framework.core.market_services import \
    MarketService, BinanceMarketService
from investing_algorithm_framework.core.models.data_provider import \
    TradingDataTypes, OrderBook, Ticker
from investing_algorithm_framework.core.models import OrderSide, Order, \
    Position, TimeUnit, db, Portfolio, OrderType, OrderStatus, \
    SQLLitePortfolio, SQLLiteOrder, SQLLitePosition
from investing_algorithm_framework.core.exceptions import \
    OperationalException, ImproperlyConfigured
from investing_algorithm_framework.app import App
from investing_algorithm_framework.globals import current_app
from investing_algorithm_framework.configuration import Config
from investing_algorithm_framework.exceptions import ApiException
from investing_algorithm_framework.views import *

VERSION = (0, 16, 7, 'alpha', 0) 
 
 
__all__ = [
    "App",
    'get_version',
    'PortfolioManager',
    'BinancePortfolioManager',
    "AlgorithmContextInitializer",
    "Portfolio",
    "OrderSide",
    "OrderType",
    "Order",
    "Position",
    "TimeUnit",
    "db",
    "current_app",
    "AlgorithmContext",
    "OrderExecutor",
    'BinanceOrderExecutor',
    "Config",
    "MarketService",
    "BinanceMarketService",
    "OrderStatus",
    "TradingDataTypes",
    "Ticker",
    "OrderBook",
    "OperationalException",
    "ImproperlyConfigured",
    "SQLLitePortfolioManager",
    "DataProvider",
    "SQLLitePortfolio",
    "SQLLiteOrder",
    "SQLLitePosition",
    "ApiException"
]



from abc import ABC, abstractmethod
from investing_algorithm_framework.core.market_identifier \
    import MarketIdentifier
from investing_algorithm_framework.core.models import TimeInterval


class MarketService(ABC, MarketIdentifier):

    @abstractmethod
    def pair_exists(self, target_symbol: str, trading_symbol: str):
        pass

    def get_price(self, target_symbol: str, trading_symbol: str):
        return self.get_ticker(target_symbol, trading_symbol)

    @abstractmethod
    def get_prices(
        self,
        target_symbol: str,
        trading_symbol: str,
        interval: TimeInterval
    ):
        pass

    @abstractmethod
    def get_ticker(self, target_symbol: str, trading_symbol: str):
        pass

    @abstractmethod
    def get_order_book(self, target_symbol: str, trading_symbol: str):
        pass

    @abstractmethod
    def get_balance(self, symbol: str = None):
        pass

    @abstractmethod
    def create_limit_buy_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
        price: float
    ):
        pass

    @abstractmethod
    def create_limit_sell_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
        price: float
    ):
        pass

    @abstractmethod
    def create_market_buy_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
    ):
        pass

    @abstractmethod
    def create_market_sell_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
    ):
        pass

    @abstractmethod
    def get_orders(self, target_symbol: str, trading_symbol: str):
        pass

    @abstractmethod
    def get_order(self, order_id, target_symbol: str, trading_symbol: str):
        pass

    @abstractmethod
    def get_open_orders(
        self, target_symbol: str = None, trading_symbol: str = None
    ):
        pass

    @abstractmethod
    def get_closed_orders(
        self, target_symbol: str = None, trading_symbol: str = None
    ):
        pass

    @abstractmethod
    def cancel_order(self, order_id):
        pass



# Framework constants
FRAMEWORK_NAME = 'FRAMEWORK_NAME'

# Database Constants
DATABASE_CONFIG = 'DATABASE_CONFIG'
DATABASE_NAME = 'DATABASE_NAME'
DATABASE_TYPE = 'DATABASE_TYPE'
DATABASE_DIRECTORY_PATH = 'DATABASE_DIRECTORY_PATH'
DATABASE_URL = 'DATABASE_URL'

RESOURCES_DIRECTORY = "RESOURCES_DIRECTORY"
LOG_LEVEL = 'LOG_LEVEL'
BASE_DIR = 'BASE_DIR'
SQLALCHEMY_DATABASE_URI = 'SQLALCHEMY_DATABASE_URI'


BINANCE = "BINANCE"

IDENTIFIER_QUERY_PARAM = "identifier"
TARGET_SYMBOL_QUERY_PARAM = "target_symbol"
TRADING_SYMBOL_QUERY_PARAM = "trading_symbol"
ORDER_SIDE_QUERY_PARAM = "side"
STATUS_QUERY_PARAM = "status"
POSITION_SYMBOL_QUERY_PARAM = "position"
SYMBOL_QUERY_PARAM = "symbol"
TIME_FRAME_QUERY_PARAM = "time_frame"

RESERVED_IDENTIFIERS = [
    BINANCE
]

BINANCE_API_KEY = "binance_api_key"
BINANCE_SECRET_KEY = "binance_secret_key"

TRADING_SYMBOL = "TRADING_SYMBOL"

CHECK_PENDING_ORDERS = "CHECK_PENDING_ORDERS"



from .portfolio_manager import PortfolioManager
from .sqllite_portfolio_manager import SQLLitePortfolioManager
from .binance import BinancePortfolioManager
from .factory import DefaultPortfolioManagerFactory


__all__ = [
    "PortfolioManager",
    "SQLLitePortfolioManager",
    "BinancePortfolioManager",
    "DefaultPortfolioManagerFactory"
]



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.mixins import BinanceOrderExecutorMixin
from investing_algorithm_framework.core.order_executors import OrderExecutor


class BinanceOrderExecutor(BinanceOrderExecutorMixin, OrderExecutor):
    identifier = BINANCE

    def __init__(self, identifier: str = None):
        super(BinanceOrderExecutor, self).__init__(identifier)

        if identifier is not None:
            self.identifier = identifier



from abc import abstractmethod

from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.identifier import Identifier
from investing_algorithm_framework.core.market_identifier import \
    MarketIdentifier
from investing_algorithm_framework.core.models import db, SQLLitePortfolio, \
    OrderStatus, OrderType, OrderSide, Portfolio, SQLLiteOrder, SQLLitePosition
from investing_algorithm_framework.core.portfolio_managers.portfolio_manager\
    import PortfolioManager


class SQLLitePortfolioManager(PortfolioManager, Identifier, MarketIdentifier):
    trading_symbol = None

    @abstractmethod
    def get_unallocated_synced(self, algorithm_context):
        pass

    @abstractmethod
    def get_positions_synced(self, algorithm_context):
        pass

    def initialize(self, algorithm_context):
        self._initialize_portfolio(algorithm_context)

    def _initialize_portfolio(self, algorithm_context):
        portfolio = self.get_portfolio(False)

        self.trading_symbol = self.get_trading_symbol(algorithm_context)

        if portfolio is None:

            unallocated = 0
            synced_unallocated = self.get_unallocated_synced(algorithm_context)

            if synced_unallocated is not None:
                unallocated = synced_unallocated

            portfolio = SQLLitePortfolio(
                identifier=self.identifier,
                trading_symbol=self.trading_symbol,
                unallocated=unallocated,
                market=self.get_market()
            )
            portfolio.save(db)
        else:
            self.get_unallocated(algorithm_context, True)
            self.get_allocated(algorithm_context, True)

    def get_unallocated(self, algorithm_context, sync=False):
        portfolio = self.get_portfolio()

        if sync:
            unallocated = self.get_unallocated_synced(algorithm_context)

            if portfolio.unallocated != unallocated:
                portfolio.unallocated = unallocated
                db.session.commit()

        return portfolio.unallocated

    def get_allocated(self, algorithm_context, sync=False):
        portfolio = self.get_portfolio()

        if sync:
            synced_positions = self.get_positions_synced(algorithm_context)
            positions = SQLLitePosition\
                .query.filter_by(portfolio=self.get_portfolio())

            if synced_positions is not None:
                for synced_position in synced_positions:

                    position = positions\
                        .filter_by(symbol=synced_position["symbol"])\
                        .first()

                    if position is None:

                        position = SQLLitePosition(
                            symbol=synced_position["symbol"]
                        )
                        position.amount = synced_position["amount"]
                        position.portfolio = portfolio
                        db.session.commit()
                    elif position.amount != synced_position["amount"]:
                        position.amount = synced_position["amount"]
                        db.session.commit()

        return portfolio.unallocated

    def get_portfolio(self, throw_exception=True) -> Portfolio:
        portfolio = SQLLitePortfolio.query\
            .filter_by(identifier=self.identifier)\
            .first()

        if portfolio is None and throw_exception:
            raise OperationalException("No portfolio model implemented")

        return portfolio

    def get_positions(self, symbol: str = None, lazy=False):

        if symbol is None:
            query_set = SQLLitePosition.query \
                .filter_by(portfolio=self.get_portfolio())
        else:
            query_set = SQLLitePosition.query \
                .filter_by(portfolio=self.get_portfolio()) \
                .filter_by(symbol=symbol)

        if lazy:
            return query_set
        else:
            return query_set.all()

    def get_orders(self, symbol: str = None, status=None, lazy=False):
        positions = SQLLitePosition.query \
            .filter_by(portfolio=self.get_portfolio()) \
            .with_entities(SQLLitePosition.id)

        query_set = SQLLiteOrder.query \
            .filter(SQLLiteOrder.position_id.in_(positions))

        if symbol is not None:
            query_set = query_set.filter_by(symbol=symbol)

        if status is not None:
            status = OrderStatus.from_value(status)
            query_set = query_set.filter_by(status=status.value)

        if lazy:
            return query_set
        else:
            return query_set.all()

    def create_order(
        self,
        symbol,
        price=None,
        amount_trading_symbol=None,
        amount_target_symbol=None,
        order_type=OrderType.LIMIT.value,
        order_side=OrderSide.BUY.value,
        context=None,
        validate_pair=True
    ):

        if context is None:
            from investing_algorithm_framework import current_app
            context = current_app.algorithm

        if validate_pair:
            market_service = context.get_market_service(self.market)
            if not market_service.pair_exists(symbol, self.trading_symbol):
                raise OperationalException(
                    f"Pair {symbol} {self.trading_symbol} does not exist "
                    f"on market {self.market}"
                )

        return self.get_portfolio().create_order(
            context=context,
            order_type=order_type,
            symbol=symbol,
            price=price,
            amount_trading_symbol=amount_trading_symbol,
            amount_target_symbol=amount_target_symbol,
            order_side=order_side,
        )

    def add_order(self, order):
        self.get_portfolio().add_order(order)



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.mixins import \
    BinancePortfolioManagerMixin
from investing_algorithm_framework.core.portfolio_managers\
    .sqllite_portfolio_manager import SQLLitePortfolioManager


class BinancePortfolioManager(
    BinancePortfolioManagerMixin, SQLLitePortfolioManager
):
    identifier = BINANCE
    market = BINANCE

    def __init__(self, identifier: str = None, trading_symbol: str = None):
        super(BinancePortfolioManager, self).__init__(identifier)

        if identifier is not None:
            self.identifier = identifier

        if trading_symbol is not None:
            self.trading_symbol = trading_symbol



from marshmallow import Schema, fields


class OrderSerializer(Schema):
    id = fields.Int(dump_only=True)
    order_reference = fields.String(dump_only=True)
    target_symbol = fields.String(dump_only=True)
    trading_symbol = fields.String(dump_only=True)
    initial_price = fields.Float(dump_only=True)
    amount_target_symbol = fields.Float(dump_only=True)
    amount_trading_symbol = fields.Float(dump_only=True)
    status = fields.String(dump_only=True)
    order_type = fields.String(dump_only=True)
    order_side = fields.String(dump_only=True)
    executed_at = fields.DateTime(dump_only=True)
    successful = fields.String(dump_only=True)

    # Optional fields
    identifier = fields.Method("get_identifier")
    position_id = fields.Int(dump_only=True)

    @staticmethod
    def get_identifier(obj):

        if obj.position is None:
            return None

        return obj.position.portfolio.identifier



class SQLAlchemyModelExtension:

    def update(self, db, data, commit=True, **kwargs):

        for attr, value in data.items():
            setattr(self, attr, value)

        if commit:
            db.session.commit()

    def save(self, db, commit=True):
        db.session.add(self)

        if commit:
            db.session.commit()

    def delete(self, db, commit=True):
        db.session.delete(self)

        if commit:
            db.session.commit()



from investing_algorithm_framework.core.market_services.market_service \
    import MarketService
from investing_algorithm_framework.core.market_services.binance_market_service \
    import BinanceMarketService
from investing_algorithm_framework.core.market_services.factory import \
    DefaultMarketServiceFactory

__all__ = [
    "MarketService",
    "BinanceMarketService",
    "DefaultMarketServiceFactory"
]



from flask_apscheduler import APScheduler

from investing_algorithm_framework.core.models import TimeUnit
from investing_algorithm_framework.utils import random_string
from investing_algorithm_framework.core.models import db


class Worker:

    def __init__(self, decorated, worker_id, time_unit, interval):
        from investing_algorithm_framework.core.context import algorithm

        self.decorated = decorated
        self.worker_id = worker_id

        if self.worker_id is None:
            self.worker_id = self.decorated.__name__

            if self.decorated.__name__ in algorithm.workers:
                self.id = random_string(10)

        if isinstance(time_unit, TimeUnit):
            self.time_unit = time_unit.value
        else:
            self.time_unit = TimeUnit.from_string(time_unit).value

        self.interval = interval
        algorithm.add_worker(self)

    def add_to_scheduler(self, app_scheduler: APScheduler):
        if TimeUnit.SECONDS.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                seconds=self.interval
            )
        elif TimeUnit.MINUTE.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                minutes=self.interval
            )
        elif TimeUnit.HOUR.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                minutes=(self.interval * 60)
            )

    def __call__(
            self,
            time_unit: TimeUnit = TimeUnit.MINUTE.value,
            interval=10
    ):
        from investing_algorithm_framework.core.context import algorithm

        # Run the decorated function in app context
        with db.app.app_context():
            self.decorated(algorithm)



from abc import abstractmethod
from random import randint

from sqlalchemy import UniqueConstraint
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, validates

from investing_algorithm_framework.core.models import db, OrderSide, \
    OrderStatus, Position
from investing_algorithm_framework.core.models.model_extension \
    import SQLAlchemyModelExtension


def random_id():
    """
    Function to create a random ID. This function checks first if
    the generated ID is not already taken.
    Returns: random integer that can be used as an ID
    """
    minimal = 100
    maximal = 1000000000000000000
    rand = randint(minimal, maximal)

    while SQLLitePosition.query.filter_by(id=rand).first() is not None:
        rand = randint(minimal, maximal)

    return rand


class SQLLitePosition(Position, db.Model, SQLAlchemyModelExtension):
    __tablename__ = "positions"

    # Integer id for the Position as the primary key
    id = db.Column(
        db.Integer, primary_key=True, unique=True, default=random_id
    )

    # Asset Symbol (e.g. BTC)
    symbol = db.Column(db.String)

    # The price of the asset
    orders = db.relationship(
        "SQLLiteOrder",
        back_populates="position",
        lazy="dynamic",
        cascade="all, delete-orphan"
    )

    amount = db.Column(db.Float)

    # Relationships
    portfolio_id = db.Column(db.Integer, db.ForeignKey('portfolios.id'))
    portfolio = relationship("SQLLitePortfolio", back_populates="positions")

    # Constraints
    __table_args__ = (
        UniqueConstraint(
            'symbol', 'portfolio_id', name='_symbol_portfolio_uc'
        ),
    )

    def __init__(self, symbol):
        self.symbol = symbol
        self.amount = 0

    @validates('id', 'symbol')
    def _write_once(self, key, value):
        existing = getattr(self, key)
        if existing is not None:
            raise ValueError("{} is write-once".format(key))
        return value

    @hybrid_property
    def delta(self):
        orders = self.orders\
            .filter_by(order_side=OrderSide.BUY.value) \
            .filter_by(status=OrderStatus.SUCCESS.value) \
            .all()

        delta = 0

        for order in orders:
            delta += order.delta
        return delta

    def get_symbol(self):
        return self.symbol

    def get_orders(self):
        return self.orders.all()

    @abstractmethod
    def get_amount(self):
        return self.amount

    def __repr__(self):
        return self.repr(
            id=self.id,
            symbol=self.symbol,
            amount=self.amount,
        )



import logging
from investing_algorithm_framework.core.exceptions import ImproperlyConfigured
from investing_algorithm_framework.configuration.constants import \
    DATABASE_CONFIG, LOG_LEVEL, RESOURCES_DIRECTORY

logger = logging.getLogger(__name__)


class ConfigValidator:
    required_variables = [
        DATABASE_CONFIG,
        RESOURCES_DIRECTORY,
        LOG_LEVEL
    ]

    @staticmethod
    def validate(config):
        logger.info("Validating application configuration")

        for variable in ConfigValidator.required_variables:

            if variable not in config or config.get(variable, None) is None:
                raise ImproperlyConfigured(
                    "{} is not set".format(variable)
                )



from abc import abstractmethod

from sqlalchemy.orm import relationship

from investing_algorithm_framework.core.models import db, OrderSide, \
    OrderStatus
from investing_algorithm_framework.core.models.model_extension \
    import SQLAlchemyModelExtension


class PositionSnapshot:

    @abstractmethod
    def get_amount(self):
        pass

    @abstractmethod
    def get_cost(self):
        pass

    @abstractmethod
    def get_symbol(self):
        pass

    @staticmethod
    @abstractmethod
    def from_position(position):
        pass


class SQLLitePositionSnapshot(
    PositionSnapshot, db.Model, SQLAlchemyModelExtension
):
    __tablename__ = "position_snapshots"

    # Integer id for the Position as the primary key
    id = db.Column(db.Integer, primary_key=True)

    # Asset Symbol (e.g. BTC)
    symbol = db.Column(db.String)

    amount = db.Column(db.Float)
    cost = db.Column(db.Float)

    # Relationships
    portfolio_id = db.Column(db.Integer, db.ForeignKey('portfolio_snapshots.id'))
    portfolio = relationship("SQLLitePortfolioSnapshot", back_populates="positions")

    def __init__(self, symbol, amount, cost):
        self.symbol = symbol
        self.amount = amount
        self.cost = cost

    @staticmethod
    def from_position(position):
        orders = position.orders\
            .filter_by(
                order_side=OrderSide.BUY.value,
                status=OrderStatus.SUCCESS.value
            )\
            .all()

        cost = 0

        # Calculate the cost
        for order in orders:
            cost += order.amount_target_symbol * order.initial_price

        # Create an algorithm position snapshot
        return SQLLitePositionSnapshot(
            symbol=position.symbol,
            amount=position.amount,
            cost=cost
        )

    def get_amount(self):
        return self.amount

    def get_cost(self):
        return self.cost

    def get_symbol(self):
        return self.symbol

    def __repr__(self):
        return self.repr(
            id=self.id,
            symbol=self.symbol,
            amount=self.amount,
        )



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.market_services\
    .binance_market_service import BinanceMarketService


class DefaultMarketServiceFactory:

    @staticmethod
    def of_market(market):
        if market.upper() == BINANCE:
            return BinanceMarketService()

        return None



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.portfolio_managers.binance import \
    BinancePortfolioManager


class DefaultPortfolioManagerFactory:

    @staticmethod
    def of_market(market):

        if market.upper() == BINANCE:
            return BinancePortfolioManager()

        return None



from investing_algorithm_framework.configuration.constants import BINANCE


class BinancePortfolioManagerMixin:

    def get_unallocated_synced(self, algorithm_context):
        market_service = algorithm_context.get_market_service(BINANCE)
        return market_service.get_balance(self.trading_symbol)

    def get_positions_synced(self, algorithm_context):
        market_service = algorithm_context.get_market_service(BINANCE)
        return market_service.get_balance()




from .exceptions import OperationalException


class MarketIdentifier:
    market = None

    def __init__(self, market: str = None):

        if self.market is None:
            self.market = market

        if self.market is None:
            raise OperationalException(
                f"{self.__class__.__name__} has no market specified"
            )

    def get_market(self, throw_exception=True) -> str:
        value = getattr(self, 'market', None)

        if value is None and throw_exception:
            raise OperationalException(
                f"{self.__class__.__name__} should either include an market "
                f"attribute, or override the `get_market()`, method."
            )

        return value



from marshmallow import Schema, fields
from investing_algorithm_framework.core.models import Position, Order, \
    PerformanceMetric, TimeFrame
from investing_algorithm_framework.core.performance import PerformanceService


class PortfolioSerializer(Schema):
    id = fields.Integer(dump_only=True)
    identifier = fields.String(dump_only=True)
    trading_symbol = fields.String(dump_only=True)
    unallocated = fields.Float(dump_only=True)
    allocated = fields.Float(dump_only=True)
    allocated_percentage = fields.Float(dump_only=True)
    unallocated_percentage = fields.Float(dump_only=True)
    realized = fields.Float(dump_only=True)
    market = fields.String(dump_only=True)
    created_at = fields.DateTime(dump_only=True)
    updated_at = fields.DateTime(dump_only=True)
    orders = fields.Method("get_orders")
    positions = fields.Method("get_positions")
    performance = fields.Method("get_performance")
    delta = fields.Method("get_delta")

    @staticmethod
    def get_orders(obj):
        return obj.get_number_of_orders()

    @staticmethod
    def get_positions(obj):
        return obj.get_number_of_positions()

    def get_performance(self, obj):
        return PerformanceService\
            .of_metric(
                obj,
                PerformanceMetric.OVERALL_PERFORMANCE,
                TimeFrame.from_value(
                    self.context.get("time_frame", TimeFrame.ONE_DAY.value)
                )
            )

    def get_delta(self, obj):
        return PerformanceService \
            .of_metric(
                obj,
                PerformanceMetric.DELTA,
                TimeFrame.from_value(
                    self.context.get("time_frame", TimeFrame.ONE_DAY.value)
                )
            )



import logging
from datetime import datetime
from random import randint

from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import validates

from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import db, OrderType, \
    OrderStatus, OrderSide, Order
from investing_algorithm_framework.core.models.model_extension \
    import SQLAlchemyModelExtension

logger = logging.getLogger(__name__)


def random_id():
    """
    Function to create a random ID. This function checks first if
    the generated ID is not already taken.
    Returns: random integer that can be used as an ID
    """
    minimal = 100
    maximal = 1000000000000000000
    rand = randint(minimal, maximal)

    while SQLLiteOrder.query.filter_by(id=rand).first() is not None:
        rand = randint(minimal, maximal)

    return rand


class SQLLiteOrder(Order, db.Model, SQLAlchemyModelExtension):
    """
        Class AlgorithmOrder: a database model for an AlgorithmOrder instance.

        Attributes:
        An AlgorithmOrder instance consists out of the following attributes:

        - id: unique identification also used externally
        - target_symbol: the target asset symbol of the order
        - trading_symbol: the asset that is traded for the target symbol
        - order side: the side of the order, e.g. BUY or SELL
        - order type: the order type, e.g LIMIT.

        The following attributes are OPTIONAL:

        - price: the price of the target symbol (in trading symbol currency)
        - amount: the amount of the target symbol

        Updated post-execution:

        for all child orders
        - completed_at: DateTime the order was marked as completed

        Relations:

        - position: relation to the AlgorithmPosition instances
        of users subscribed to the algorithm
        """

    __tablename__ = "orders"

    # Integer id for the Order as the primary key
    id = db.Column(
        db.Integer,
        primary_key=True,
        unique=True,
        default=random_id
    )
    order_reference = db.Column(db.Integer)
    target_symbol = db.Column(db.String)
    trading_symbol = db.Column(db.String)

    order_side = db.Column(
        db.String,
        nullable=False,
        default=OrderSide.BUY.value
    )
    order_type = db.Column(
        db.String,
        nullable=False,
        default=OrderType.LIMIT.value
    )

    # The price and amount of the asset
    initial_price = db.Column(db.Float)
    closing_price = db.Column(db.Float)
    amount = db.Column(db.Float)
    amount_trading_symbol = db.Column(db.Float)
    amount_target_symbol = db.Column(db.Float)
    status = db.Column(db.String)

    position_id = db.Column(
        db.Integer, db.ForeignKey('positions.id')
    )
    position = db.relationship("SQLLitePosition", back_populates="orders")

    # Standard date time attributes
    created_at = db.Column(db.DateTime, default=datetime.utcnow)
    updated_at = db.Column(
        db.DateTime,
        default=datetime.utcnow,
        onupdate=datetime.utcnow
    )

    # Post execution
    executed_at = db.Column(db.DateTime)

    def __init__(
        self,
        order_side,
        order_type,
        target_symbol,
        trading_symbol,
        price=None,
        amount_target_symbol=None,
        amount_trading_symbol=None,
        **kwargs
    ):
        self.order_side = OrderSide.from_value(order_side).value
        self.order_type = OrderType.from_value(order_type).value
        self.target_symbol = target_symbol
        self.trading_symbol = trading_symbol
        self.status = None
        self.initial_price = price

        if OrderType.MARKET.equals(self.order_type):

            if OrderSide.SELL.equals(self.order_side):
                if amount_target_symbol is None:
                    raise OperationalException(
                        "Amount target symbol needs to be set "
                        "for a market sell order"
                    )

                self.set_amount_target_symbol(amount_target_symbol)
            else:
                raise OperationalException(
                    "Market buy order is not yet supported"
                )
        else:
            if amount_target_symbol is None \
                    and amount_trading_symbol is None:
                raise OperationalException(
                    "Amount target symbol or amount trading "
                    "symbol needs to be set for limit order"
                )

            if amount_target_symbol is not None:
                self.set_amount_target_symbol(amount_target_symbol)
            elif amount_trading_symbol is not None:
                self.set_amount_trading_symbol(amount_trading_symbol)

        super(Order, self).__init__(**kwargs)

    def set_amount_target_symbol(self, amount):
        self.amount_target_symbol = amount

        if OrderType.LIMIT.equals(self.order_type):
            self.amount_trading_symbol = \
                self.initial_price * self.amount_target_symbol

    def set_amount_trading_symbol(self, amount):
        self.amount_trading_symbol = amount

        if OrderType.LIMIT.equals(self.order_type):
            self.amount_target_symbol = \
                self.amount_trading_symbol / self.initial_price

    @validates(
        'id',
        'target_symbol',
        'trading_symbol',
        'order_side',
        'order_type',
    )
    def _write_once(self, key, value):
        existing = getattr(self, key)

        if existing is not None:
            raise ValueError("{} is write-once".format(key))

        return value

    @property
    def current_price(self):
        from investing_algorithm_framework import current_app
        market_service = current_app.algorithm \
            .get_market_service(self.position.portfolio.market)

        current_price = self.initial_price

        try:
            current_price = market_service.get_price(
                target_symbol=self.target_symbol,
                trading_symbol=self.trading_symbol
            )
        except Exception as e:
            logger.error(e)
            return current_price

        return current_price.price

    @hybrid_property
    def delta(self):

        # Delta is 0 if it is a sell order
        if OrderSide.SELL.equals(self.order_side):
            return 0

        if self.closing_price:
            return (self.closing_price * self.amount_target_symbol) - \
                   (self.initial_price * self.amount_target_symbol)

        if not OrderStatus.SUCCESS.equals(self.status):
            return 0

        price = self.current_price

        # With no price data_provider return 0
        if price is None or price == -1:
            return 0

        return (price * self.amount_target_symbol) - \
               (self.initial_price * self.amount_target_symbol)

    @hybrid_property
    def percentage(self):

        if OrderSide.SELL.equals(self.order_side):
            return 0

        if not OrderStatus.SUCCESS.equals(self.status):
            return 0

        portfolio = self.position.portfolio

        return (
           self.current_value /
           (portfolio.allocated + portfolio.unallocated)
        ) * 100

    @hybrid_property
    def percentage_position(self):

        if OrderSide.SELL.equals(self.order_side):
            return 0

        if not OrderStatus.SUCCESS.equals(self.status):
            return 0

        return self.amount_target_symbol / self.position.amount * 100

    @hybrid_property
    def percentage_realized(self):

        if OrderSide.SELL.equals(self.order_side):
            return 0

        if not OrderStatus.CLOSED.equals(self.status):
            return 0

        portfolio = self.position.portfolio

        return (self.amount_target_symbol * self.closing_price) \
                / portfolio.realized * 100

    @hybrid_property
    def current_value(self):

        # Current value is 0 if it is a sell order
        if OrderSide.SELL.equals(self.order_side):
            return 0

        if self.status is None or OrderStatus.CANCELED.equals(self.status):
            return 0

        # Order is closed
        if OrderStatus.CLOSED.equals(self.status):
            return self.closing_price * self.amount_target_symbol

        if OrderStatus.PENDING.equals(self.status):
            return self.price * self.amount_target_symbol

        price = self.current_price
        return price * self.amount_target_symbol

    def set_executed(
        self, snapshot=True, executed_at=None, amount=None, price=None
    ):

        # Can't execute limit orders that have 0 size
        if OrderType.LIMIT.equals(self.order_type) \
                and self.amount_target_symbol == 0:
            return

        if not OrderStatus.PENDING.equals(self.status) and \
                self.status is not None:
            raise OperationalException("Order is not in pending state")

        if self.amount_target_symbol != 0 and self.position is None:
            raise OperationalException("Order is not linked to a position")

        if OrderType.MARKET.equals(self.order_type):

            if amount is None or price is None:
                raise OperationalException(
                    "Amount and price needs to be provided for execution "
                    "of a market order"
                )
            self.initial_price = price

            if OrderSide.SELL.equals(self.order_side):
                self.set_amount_trading_symbol(amount)

        if executed_at is not None:
            self.executed_at = executed_at
        else:
            self.executed_at = datetime.utcnow()

        if OrderSide.BUY.equals(self.order_side):
            self.status = OrderStatus.SUCCESS.value
            self.position.amount += self.amount_target_symbol
            self.position.portfolio.total_cost += \
                self.amount_target_symbol * self.initial_price
        else:
            self.status = OrderStatus.SUCCESS.value
            self.close_buy_orders()
            self.position.portfolio.total_revenue += \
                self.amount_target_symbol * self.initial_price

        db.session.commit()

        if snapshot:
            self.snapshot(self.executed_at)

    def set_closed(self, sell_order):
        self.status = OrderStatus.CLOSED.value
        self.closing_price = sell_order.initial_price

        if OrderSide.BUY.equals(self.order_side):
            self.position.portfolio.unallocated += self.current_value
            self.position.portfolio.realized += self.delta
            self.position.amount -= self.amount_target_symbol

        db.session.commit()

    def set_pending(self):

        if not OrderStatus.TO_BE_SENT.equals(self.status):
            raise OperationalException("Order status is not TO_BE_SENT")

        self.status = OrderStatus.PENDING.value
        db.session.commit()

    def update(self, db, data, commit=True, **kwargs):
        self.updated_at = datetime.utcnow()
        super(Order, self).update(db, data, commit, **kwargs)

    def copy(self, amount=None):

        if amount is None:
            amount = self.amount

        if amount > self.amount_target_symbol:
            raise OperationalException("Amount is larger then original order")

        order = SQLLiteOrder(
            amount_trading_symbol=self.get_amount_trading_symbol(),
            price=self.initial_price,
            order_side=self.order_side,
            order_type=self.order_type,
            target_symbol=self.target_symbol,
            trading_symbol=self.trading_symbol,
            amount_target_symbol=self.amount_target_symbol - (
                    self.amount_target_symbol - amount)
        )

        order.amount_trading_symbol = order.initial_price \
                                      * order.amount_target_symbol
        order.order_reference = self.order_reference
        order.status = self.status
        order.executed_at = self.executed_at
        order.updated_at = self.updated_at
        order.created_at = self.created_at
        order.closing_price = self.closing_price

        self.amount_trading_symbol -= order.amount_trading_symbol
        self.amount_target_symbol -= order.amount_target_symbol

        return order

    def split(self, amount):

        if not OrderSide.BUY.equals(self.order_side):
            raise OperationalException("Sell order can't be split")

        if not OrderStatus.SUCCESS.equals(self.status):
            raise OperationalException("Order can't be split")

        if amount <= 0 or amount >= self.amount_target_symbol:
            raise OperationalException("Split amount has a wrong value")

        algorithm_order = self.copy(amount=amount)
        self.position.orders.append(algorithm_order)
        db.session.commit()

        return self, algorithm_order

    def save(self, db, commit=True):

        if self.position is None:
            raise OperationalException(
                "Can't save order that is not linked to an position"
            )

        super(Order, self).save(db, commit)

    def cancel(self, commit=True):
        from investing_algorithm_framework import current_app

        if OrderStatus.PENDING.equals(self.status) or \
                OrderStatus.TO_BE_SENT.equals(self.status) or \
                self.status is None:

            market_service = current_app.algorithm \
                .get_market_service(self.position.portfolio.market)

            market_service.cancel_order(self.order_reference)

            self.update(
                db, {"status": OrderStatus.CANCELED.value}, commit=commit
            )

            if OrderSide.BUY.equals(self.order_side):
                self.position.portfolio.unallocated += \
                    self.amount_trading_symbol

            if commit:
                db.session.commit()

            self.snapshot()

    def snapshot(self, creation_datetime=None):
        from investing_algorithm_framework.core.models.snapshots \
            import SQLLitePortfolioSnapshot
        SQLLitePortfolioSnapshot \
            .from_portfolio(
                self.position.portfolio,
                creation_datetime=creation_datetime
            ) \
            .save(db)

    def close_buy_orders(self):

        if OrderSide.SELL.equals(self.order_side) \
                and OrderStatus.SUCCESS.equals(self.status):
            sell_amount = self.amount_target_symbol

            if sell_amount != 0:
                # Close open buy orders
                buy_orders = self.position.orders \
                    .filter_by(status=OrderStatus.SUCCESS.value) \
                    .filter_by(order_side=OrderSide.BUY.value) \
                    .order_by(SQLLiteOrder.created_at.desc()) \
                    .all()

                index = 0

                while sell_amount != 0:
                    buy_order = buy_orders[index]

                    if buy_order.amount_target_symbol <= sell_amount:
                        sell_amount -= buy_order.amount_target_symbol
                        buy_order.set_closed(self)
                    else:
                        remainder = buy_order.amount_target_symbol - \
                                    sell_amount
                        og, split = buy_order.split(remainder)
                        og.set_closed(self)
                        sell_amount = 0

                    index += 1

    def get_id(self):
        return self.id

    def get_order_reference(self):
        return self.order_reference

    def set_order_reference(self, order_reference):
        self.order_reference = order_reference

    def get_target_symbol(self):
        return self.target_symbol

    def get_trading_symbol(self):
        return self.trading_symbol

    def get_initial_price(self):
        return self.initial_price

    def set_initial_price(self, price):
        self.initial_price = price

    def get_price(self):
        return self.current_price

    def get_closing_price(self):
        return self.closing_price

    def set_closing_price(self, price):
        self.closing_price = price

    def get_side(self):
        return self.order_side

    def get_status(self):
        return self.status

    def set_status(self, status: OrderStatus):
        self.status = OrderStatus.from_value(status).value

    def get_type(self):
        return self.order_type

    def get_amount_trading_symbol(self):
        return self.amount_trading_symbol

    def get_amount_target_symbol(self):
        return self.amount_target_symbol

    @staticmethod
    def from_dict(data):
        pass

    def to_dict(self):
        pass

    def __repr__(self):
        return self.to_string()


from .exceptions import OperationalException


class Identifier:
    identifier = None
    RESERVED = [
        "BINANCE"
    ]

    def __init__(self, identifier: str = None, throw_exception=True):

        if self.identifier is None:
            self.identifier = identifier

        # If ID is none generate a new unique ID
        if self.identifier is None and throw_exception:
            raise OperationalException(
                f"{self.__class__.__name__} has no identifier specified"
            )

    def get_id(self, throw_exception=True) -> str:
        value = getattr(self, 'identifier', None)

        if value is None and throw_exception:
            raise OperationalException(
                f"{self.__class__.__name__} should either include an "
                f"identifier attribute, or override the `get_id()`, method."
            )

        return getattr(self, 'identifier')



from investing_algorithm_framework.core.models.data_provider import \
    OrderBook, Ticker
from investing_algorithm_framework.core.identifier import Identifier
from investing_algorithm_framework.core.exceptions import OperationalException


class DataProvider(Identifier):

    def provide_raw_data(self, algorithm_context, **kwargs) -> dict:
        raise OperationalException(
            f"Data provider {self.__class__.__name__} does not support "
            f"trading data type raw data, please specify a supported "
            f"trading data type."
        )

    def provide_ticker(
            self, target_symbol, trading_symbol, algorithm_context, **kwargs
    ) -> Ticker:
        raise OperationalException(
            f"Data provider {self.__class__.__name__} does not support "
            f"trading data type ticker, please specify a supported "
            f"trading data type."
        )

    def provide_order_book(
            self, target_symbol, trading_symbol, algorithm_context, **kwargs
    ) -> OrderBook:
        raise OperationalException(
            f"Data provider {self.__class__.__name__} does not support "
            f"trading data type order book, please specify a supported "
            f"trading data type."
        )



from investing_algorithm_framework.core.order_validators.order_validator import OrderValidator
from investing_algorithm_framework.core.order_validators.order_validation_factory \
    import OrderValidatorFactory



from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.configuration.constants import \
    BINANCE_API_KEY, BINANCE_SECRET_KEY


class BinanceApiSecretKeySpecifierMixin:
    binance_api_key = None
    binance_secret_key = None

    def get_secret_key(self, throw_exception=True):
        secret_key = getattr(self, BINANCE_SECRET_KEY, None)

        if secret_key is None and throw_exception:
            raise OperationalException(
                f"Binance secret key is not set on class "
                f"{self.__class__.__name__}. "
                f"Either override 'get_secret_key' method or set "
                f"the 'secret_key' attribute in the algorithm config."
            )

        return secret_key

    def get_api_key(self, throw_exception=True):
        api_key = getattr(self, BINANCE_API_KEY, None)

        if api_key is None and throw_exception:
            raise OperationalException(
                "Binance api key is not set. Either override 'get_api_key' "
                "method or set the 'binance_api_key' attribute in the "
                "algorithm config."
            )

        return api_key



from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import Order, OrderType, \
    OrderSide, OrderStatus
from investing_algorithm_framework.configuration.constants import BINANCE


class BinanceOrderExecutorMixin:
    identifier = BINANCE

    def execute_limit_order(self, order: Order, algorithm_context, **kwargs):
        market_service = algorithm_context.get_market_service(BINANCE)

        if not OrderType.LIMIT.equals(order.order_type):
            raise OperationalException(
                "Provided order is not a limit order type"
            )

        if OrderSide.BUY.equals(order.order_side):
            market_service.create_limit_buy_order(
                target_symbol=order.target_symbol,
                trading_symbol=order.trading_symbol,
                amount=order.amount,
                price=order.price
            )
        else:
            market_service.create_limit_sell_order(
                target_symbol=order.target_symbol,
                trading_symbol=order.trading_symbol,
                amount=order.amount,
                price=order.price
            )

    def execute_market_order(self, order: Order, algorithm_context, **kwargs):
        market_service = algorithm_context.get_market_service(BINANCE)

        if not OrderType.MARKET.equals(order.order_type):
            raise OperationalException(
                "Provided order is not a market order type"
            )

        if OrderSide.BUY.equals(order.order_side):
            market_service.create_market_buy_order(
                target_symbol=order.target_symbol,
                trading_symbol=order.trading_symbol,
                amount=order.amount,
            )
        else:
            market_service.create_market_sell_order(
                target_symbol=order.target_symbol,
                trading_symbol=order.trading_symbol,
                amount=order.amount,
            )

    def get_order_status(self, order: Order, algorithm_context, **kwargs):
        market_service = algorithm_context.get_market_service(BINANCE)

        order = market_service.get_order(
            order.exchange_id, order.target_symbol, order.trading_symbol
        )

        if order is not None:

            if order["info"]["status"] == "FILLED":
                return OrderStatus.SUCCESS.value

            if order["info"]["status"] == "REJECTED	":
                return OrderStatus.FAILED.value

            if order["info"]["status"] == "PENDING_CANCEL":
                return OrderStatus.FAILED.value

            if order["info"]["status"] == "EXPIRED":
                return OrderStatus.FAILED.value

            return OrderStatus.PENDING.value



from abc import abstractmethod
from typing import List

from investing_algorithm_framework.core.models import OrderSide, \
    OrderStatus
from investing_algorithm_framework.core.models.position import Position
from investing_algorithm_framework.core.models.order import Order


class Portfolio:

    @abstractmethod
    def get_id(self):
        pass

    @abstractmethod
    def get_identifier(self):
        pass

    @abstractmethod
    def get_trading_symbol(self):
        pass

    @abstractmethod
    def get_positions(self, symbol: str = None) -> List[Position]:
        pass

    @abstractmethod
    def get_number_of_positions(self):
        pass

    @abstractmethod
    def get_orders(
        self,
        status: OrderStatus = None,
        side: OrderSide = None,
        target_symbol: str = None,
        trading_symbol: str = None,
        lazy: bool = False
    ) -> List[Order]:
        pass

    @abstractmethod
    def get_number_of_orders(self):
        pass

    def get_market(self) -> str:
        pass

    @staticmethod
    @abstractmethod
    def from_dict(data):
        pass

    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def get_unallocated(self):
        pass

    @abstractmethod
    def get_allocated(self):
        pass

    @abstractmethod
    def get_realized(self):
        pass

    @abstractmethod
    def get_total_revenue(self):
        pass

    @abstractmethod
    def snapshot(
        self, withdrawel=0, deposit=0, commit=True, creation_datetime=None
    ):
        pass

    @abstractmethod
    def create_order(
        self,
        context,
        order_type,
        symbol,
        price=None,
        amount_trading_symbol=None,
        amount_target_symbol=None,
        order_side=OrderSide.BUY.value,
    ):
        pass

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def to_string(self):
        return self.repr(
            id=self.get_id(),
            identifier=self.get_identifier(),
            trading_symbol=self.get_trading_symbol(),
            number_of_positions=self.get_number_of_positions(),
            number_of_orders=self.get_number_of_orders(),
            unallocated=f"{self.get_unallocated()} {self.get_trading_symbol()}",
            allocated=self.get_allocated(),
            realized=f"{self.get_realized()}",
            total_revenue=f"{self.get_total_revenue()}",
        )

    def __repr__(self):
        return self.to_string()



import logging
from typing import List
import inspect

from investing_algorithm_framework.configuration import Config
from investing_algorithm_framework.configuration.constants import \
    TRADING_SYMBOL
from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import TimeUnit, OrderType, \
    db, OrderSide, OrderStatus, Portfolio, Order, Position
from investing_algorithm_framework.core.models.data_provider import \
    TradingDataTypes
from investing_algorithm_framework.core.workers import Worker, Strategy
from investing_algorithm_framework.extensions import scheduler
from investing_algorithm_framework.core.data_providers import\
    DefaultDataProviderFactory, DataProvider
from investing_algorithm_framework.configuration.constants import \
    RESERVED_IDENTIFIERS
from investing_algorithm_framework.core.market_services import \
    DefaultMarketServiceFactory
from investing_algorithm_framework.core.portfolio_managers import \
    DefaultPortfolioManagerFactory
from investing_algorithm_framework.core.order_executors import \
    DefaultOrderExecutorFactory, OrderExecutor

logger = logging.getLogger(__name__)


class AlgorithmContext:
    """
    The AlgorithmContext defines the context of an algorithm.

    An algorithm consist out of an data_provider provider and a set of
    strategies that belong to the data_provider provider.
    """
    _config = None
    _workers = []
    _running_workers = []
    _data_providers = {}
    _order_executors = {}
    _portfolio_managers = {}
    _market_services = {}
    _initializer = None
    _initialized = False

    # Decorator for worker to allow for deferred calling
    @staticmethod
    def schedule(
        function=None,
        worker_id: str = None,
        time_unit: TimeUnit = TimeUnit.MINUTE,
        interval=10
    ):

        if function:
            return Worker(function, worker_id, time_unit, interval)
        else:
            def wrapper(f):
                return Worker(f, worker_id, time_unit, interval)
            return wrapper

    # Decorator for strategy worker to allow for deferred calling
    @staticmethod
    def strategy(
        function=None,
        worker_id: str = None,
        time_unit: TimeUnit = TimeUnit.MINUTE,
        interval=10,
        data_provider_identifier=None,
        trading_data_type=None,
        trading_data_types=None,
        target_symbol=None,
        target_symbols=None,
        trading_symbol=None
    ):

        if function:
            return Strategy(
                function,
                worker_id,
                time_unit,
                interval,
                data_provider_identifier,
                trading_data_type,
                trading_data_types,
                target_symbol,
                target_symbols,
                trading_symbol
            )
        else:
            def wrapper(f):
                return Strategy(
                    f,
                    worker_id,
                    time_unit,
                    interval,
                    data_provider_identifier,
                    trading_data_type,
                    trading_data_types,
                    target_symbol,
                    target_symbols,
                    trading_symbol
                )

            return wrapper

    def initialize(self, config=None):

        if config is not None:
            assert isinstance(config, Config), (
                "Config is not an instance of config"
            )
            self._config = config
        else:
            self._config = Config()

    def start(self):

        # Initialize the algorithm context
        if not self._initialized:

            # Run the initializer
            if self._initializer is not None:
                self._initializer.initialize(self)

        # Initialize the portfolio managers
        for portfolio_manager_key in self._portfolio_managers:
            portfolio_manager = self._portfolio_managers[portfolio_manager_key]
            portfolio_manager.initialize(self)

        for order_executor_key in self._order_executors:
            order_executor = self._order_executors[order_executor_key]
            order_executor.initialize(self)

        self._initialized = True

        # Start the workers
        self.start_workers()

    def stop(self):
        self.stop_all_workers()

    def start_workers(self):

        if not self.running:
            # Start functional workers
            for worker in self._workers:
                worker.add_to_scheduler(scheduler)
                self._running_workers.append(worker)

    def start_strategies(self):

        if not self.running:
            # Start functional workers
            for worker in self._workers:
                worker.add_to_scheduler(scheduler)
                self._running_workers.append(worker)

    def stop_all_workers(self):
        scheduler.remove_all_jobs()
        self._running_workers = []

    def add_worker(self, worker):
        """
        Function to ad an worker to list of workers of the
        algorithm context.
        """

        assert isinstance(worker, Worker), OperationalException(
            "Worker is not an instance of an Worker"
        )

        for installed_worker in self._workers:

            if installed_worker.worker_id == worker.worker_id:
                return

        self._workers.append(worker)

    def add_strategy(self, strategy):

        if inspect.isclass(strategy):
            strategy = strategy()

        assert isinstance(strategy, Strategy), \
            OperationalException(
                "Strategy is not an instance of a Strategy"
            )

        for installed_worker in self._workers:

            if installed_worker.worker_id == strategy.worker_id:
                return

        self._workers.append(strategy)

    @property
    def running(self) -> bool:
        """
            An utility property to check if there are active workers for the
            algorithm.
        """

        return len(self._running_workers) != 0

    @property
    def workers(self) -> List:
        return self._workers

    @property
    def running_workers(self) -> List:
        return self._running_workers

    @property
    def running_strategies(self) -> List:
        strategies = []

        for worker in self._workers:
            if isinstance(worker, Strategy):
                strategies.append(worker)

        return strategies

    @property
    def config(self) -> Config:
        return self._config

    def add_order_executor(self, order_executor):
        from investing_algorithm_framework.core.order_executors \
            import OrderExecutor

        if inspect.isclass(order_executor):
            order_executor = order_executor()

        if order_executor in RESERVED_IDENTIFIERS:
            raise OperationalException(
                "Identifier of order executor is reserved"
            )

        assert isinstance(order_executor, OrderExecutor), (
            'Provided object must be an instance of the OrderExecutor class'
        )

        if order_executor.identifier in self._order_executors:
            raise OperationalException("Order executor id already exists")

        self._order_executors[order_executor.identifier] = order_executor

    @property
    def order_executors(self) -> List:
        order_executors = []

        for order_executor in self._order_executors:
            order_executors.append(order_executor)

        return order_executors

    def get_order_executor(
        self, identifier=None, throw_exception: bool = True
    ) -> OrderExecutor:

        if identifier is None:

            if len(self._order_executors.keys()) == 0:
                raise OperationalException(
                    "Algorithm has no order executors registered"
                )

            return self._order_executors[
                list(self._order_executors.keys())[0]
            ]

        if identifier not in self._order_executors:

            if identifier in RESERVED_IDENTIFIERS:
                order_executor = DefaultOrderExecutorFactory\
                    .of_market(identifier)
                order_executor.initialize(self)
                self._order_executors[order_executor.identifier] = \
                    order_executor
                return order_executor

            if throw_exception:
                raise OperationalException(
                    f"No corresponding order executor found for "
                    f"identifier {identifier}"
                )

            return None

        return self._order_executors[identifier]

    def add_data_provider(self, data_provider):
        from investing_algorithm_framework.core.data_providers \
            import DataProvider

        if inspect.isclass(data_provider):
            data_provider = data_provider()

        if data_provider.identifier in RESERVED_IDENTIFIERS:
            raise OperationalException(
                "Identifier of data provider is reserved"
            )

        assert isinstance(data_provider, DataProvider), (
            'Provided object must be an instance of the DataProvider class'
        )

        if data_provider.identifier in self._data_providers:
            raise OperationalException("DataProvider id already exists")

        self._data_providers[data_provider.identifier] = data_provider

    @property
    def data_providers(self) -> List:
        data_providers = []

        for data_provider in self._data_providers:
            data_providers.append(data_provider)

        return data_providers

    def get_data_provider(
            self, identifier=None, throw_exception: bool = True
    ) -> DataProvider:

        if identifier is None:

            if len(self._data_providers.keys()) == 0:
                raise OperationalException(
                    "Algorithm has no data providers registered"
                )

            return self._data_providers[
                list(self._data_providers.keys())[0]
            ]

        if identifier not in self._data_providers:

            if identifier in RESERVED_IDENTIFIERS:
                data_provider = DefaultDataProviderFactory\
                    .of_identifier(identifier)

                if data_provider is not None:
                    self._data_providers[identifier.upper()] = data_provider
                    return data_provider

            if throw_exception:
                raise OperationalException(
                    f"No corresponding data provider found for "
                    f"identifier {identifier}"
                )

            return None

        return self._data_providers[identifier]

    def add_initializer(self, initializer):
        from investing_algorithm_framework.core.context \
            import AlgorithmContextInitializer

        if inspect.isclass(initializer):
            initializer = initializer()

        assert isinstance(initializer, AlgorithmContextInitializer), (
            'Provided object must be an instance of the '
            'AlgorithmContextInitializer class'
        )

        self._initializer = initializer

    def add_portfolio_manager(self, portfolio_manager):
        # Check if order executor is instance of AbstractPortfolioManager
        from investing_algorithm_framework.core.portfolio_managers \
            import PortfolioManager

        if inspect.isclass(portfolio_manager):
            portfolio_manager = portfolio_manager()

        assert isinstance(portfolio_manager, PortfolioManager), (
            'Provided object must be an instance of the '
            'AbstractPortfolioManager class'
        )

        if portfolio_manager.identifier in self._portfolio_managers:
            raise OperationalException(
                f"Portfolio manager identifier {portfolio_manager.identifier} "
                f"already exists"
            )

        self._portfolio_managers[portfolio_manager.identifier] \
            = portfolio_manager

    @property
    def portfolio_managers(self) -> List:
        portfolio_managers = []

        for key in self._portfolio_managers:
            portfolio_managers.append(self._portfolio_managers[key])

        return portfolio_managers

    def get_portfolio_manager(
        self, identifier: str = None, throw_exception: bool = True
    ):

        if identifier is None:

            if len(self._portfolio_managers.keys()) == 0:
                raise OperationalException(
                    "Algorithm has no portfolio managers registered"
                )

            return self._portfolio_managers[
                list(self._portfolio_managers.keys())[0]
            ]

        if identifier not in self._portfolio_managers:

            if identifier in RESERVED_IDENTIFIERS:
                portfolio_manager = DefaultPortfolioManagerFactory\
                    .of_market(identifier)

                portfolio_manager.initialize(self)
                self._portfolio_managers[portfolio_manager.identifier] \
                    = portfolio_manager

                return portfolio_manager

            if throw_exception:
                raise OperationalException(
                    f"No corresponding portfolio manager found for "
                    f"identifier {identifier}"
                )

            return None

        return self._portfolio_managers[identifier]

    def add_market_service(self, market_service):
        from investing_algorithm_framework.core.market_services \
            import MarketService

        if inspect.isclass(market_service):
            market_service = market_service()

        assert isinstance(market_service, MarketService), (
            'Provided object must be an instance of the MarketService class'
        )

        if market_service.market in self._market_services:
            raise OperationalException("Market service market already exists")

        self._market_services[market_service.market] = market_service

    @property
    def market_services(self) -> List:
        market_services = []

        for market_service in self._market_services:
            market_services.append(market_service)

        return market_services

    def get_market_service(
            self, market: str, throw_exception: bool = True
    ):
        if market not in self._market_services:

            if market in RESERVED_IDENTIFIERS:
                market_service = DefaultMarketServiceFactory\
                    .of_market(market)

                if market_service is not None:
                    market_service.initialize(self.config)
                    self._market_services[market.upper()] = market_service
                    return market_service

            if throw_exception:
                raise OperationalException(
                    f"No corresponding market service found for "
                    f"market {market}"
                )

            return None

        return self._market_services[market]

    def set_algorithm_context_initializer(
            self, algorithm_context_initializer
    ) -> None:

        # Check if initializer is an instance of AlgorithmContextInitializer
        from investing_algorithm_framework.core.context import \
            AlgorithmContextInitializer

        assert isinstance(
            algorithm_context_initializer, AlgorithmContextInitializer
        ), (
            'Initializer must be an instance of AlgorithmContextInitializer'
        )
        self._initializer = algorithm_context_initializer

    def deposit(self, amount, identifier: str = None):
        portfolio = self.get_portfolio_manager(identifier).get_portfolio()
        portfolio.deposit(amount)

    def withdraw(self, amount, identifier: str = None):
        portfolio = self.get_portfolio_manager(identifier).get_portfolio()
        portfolio.withdraw(amount)

    def add_order(self, order, identifier: str = None):
        portfolio_manager = self.get_portfolio_manager(identifier)
        portfolio_manager.add_order(order)

    def create_limit_buy_order(
            self,
            symbol: str,
            price: float,
            amount: float,
            execute=False,
            identifier: str = None,
            validate_pair=True
    ):
        portfolio_manager = self.get_portfolio_manager(identifier)
        order = portfolio_manager.create_order(
            symbol=symbol,
            price=price,
            amount_target_symbol=amount,
            order_type=OrderType.LIMIT.value,
            validate_pair=validate_pair
        )

        if execute:
            portfolio_manager.add_order(order)
            self.execute_limit_buy_order(identifier, order)
            order.set_pending()
            db.session.commit()

        return order

    def create_limit_sell_order(
            self,
            symbol,
            price,
            amount,
            execute=False,
            identifier: str = None,
            validate_pair=True
    ):
        portfolio_manager = self.get_portfolio_manager(identifier)
        order = portfolio_manager.create_order(
            symbol=symbol,
            price=price,
            amount_target_symbol=amount,
            order_type=OrderType.LIMIT.value,
            order_side=OrderSide.SELL.value,
            validate_pair=validate_pair
        )

        if execute:
            # Execute order and set to pending state
            portfolio_manager.add_order(order)
            self.execute_limit_sell_order(identifier, order)
            order.set_pending()

        return order

    def create_market_sell_order(
        self,
        symbol,
        amount_target_symbol,
        execute=False,
        identifier: str = None,
        validate_pair=True
    ):
        portfolio_manager = self.get_portfolio_manager(identifier)
        order = portfolio_manager.create_order(
            symbol=symbol,
            amount_target_symbol=amount_target_symbol,
            order_type=OrderType.MARKET.value,
            order_side=OrderSide.SELL.value,
            validate_pair=validate_pair
        )

        if execute:
            portfolio_manager.add_order(order)
            self.execute_market_sell_order(identifier, order)
            order.set_pending()

        return order

    def execute_limit_sell_order(self, identifier, order):
        order_executor = self.get_order_executor(identifier)
        order_executor.execute_limit_order(order, self)

    def execute_limit_buy_order(self, identifier, order):
        order_executor = self.get_order_executor(identifier)
        order_executor.execute_limit_order(order, self)

    def execute_market_sell_order(self, identifier, order):
        order_executor = self.get_order_executor(identifier)
        order_executor.execute_market_order(order, self)

    def execute_market_buy_order(self, identifier, order):
        order_executor = self.get_order_executor(identifier)
        order_executor.execute_market_order(order, self)

    def check_order_status(self, identifier=None, symbol: str = None):
        portfolio_manager = self.get_portfolio_manager(identifier)
        orders = portfolio_manager\
            .get_orders(symbol, status=OrderStatus.PENDING)
        order_executor = self.get_order_executor(identifier)

        for order in orders:
            status = order_executor.get_order_status(order, self)
            order.update(db, {"status": status}, True)

    def check_pending_orders(self, identifier=None, symbol: str = None):

        if identifier is not None:

            portfolio_manager = self.get_portfolio_manager(identifier)

            order_executor = \
                self.get_order_executor(portfolio_manager.identifier)

            pending_orders = portfolio_manager\
                .get_orders(symbol, status=OrderStatus.PENDING.value)

            for pending_order in pending_orders:
                status = order_executor.get_order_status(
                    pending_order, self
                )

                if OrderStatus.SUCCESS.equals(status):
                    pending_order.set_executed()
                else:
                    pending_order.update(db, {status: status.value})
        else:

            for portfolio_manager_key in self._portfolio_managers:
                portfolio_manager = self._portfolio_managers[
                    portfolio_manager_key
                ]

                pending_orders = portfolio_manager.get_orders(
                    symbol, status=OrderStatus.PENDING.value
                )

                order_executor = self.get_order_executor(
                    portfolio_manager.identifier
                )

                for pending_order in pending_orders:
                    status = order_executor.get_order_status(
                        pending_order, self
                    )

                    if OrderStatus.SUCCESS.equals(status):
                        pending_order.set_executed()
                    else:
                        pending_order.update(db, {status: status.value})

    def get_data(
        self,
        data_provider_identifier,
        trading_data_type=None,
        trading_data_types=None,
        target_symbol=None,
        trading_symbol=None,
        target_symbols: List = None
    ):
        data_provider = self.get_data_provider(data_provider_identifier)
        data = {}

        # Check if trading symbol is specified
        if trading_symbol is None:
            trading_symbol = self.config.get(TRADING_SYMBOL, None)

            if trading_symbol is None:
                raise OperationalException(
                    "Trading symbol is not set. Either provide a "
                    "'trading_symbol' param in the method or set "
                    "the 'trading_symbol' attribute in the algorithm "
                    "config."
                )

        if target_symbols is None and target_symbol is None \
                and (trading_data_types is not None
                     or trading_data_type is not None):
            raise OperationalException(
                "You must either set the target symbol param or "
                "target symbols param for this trading data type"
            )

        if trading_data_type is not None:

            if TradingDataTypes.TICKER.equals(trading_data_type):

                if target_symbols is not None:
                    tickers = []

                    for target_symbol in target_symbols:
                        tickers.append(
                            data_provider.provide_ticker(
                                target_symbol=target_symbol,
                                trading_symbol=trading_symbol,
                                algorithm_context=self
                            )
                        )
                    data["tickers"] = tickers
                else:
                    data["ticker"] = data_provider.provide_ticker(
                        target_symbol=target_symbol,
                        trading_symbol=trading_symbol,
                        algorithm_context=self
                    )

            elif TradingDataTypes.ORDER_BOOK.equals(trading_data_type):

                if target_symbols is not None:
                    order_books = []

                    for target_symbol in target_symbols:
                        order_books.append(
                            data_provider.provide_order_book(
                                target_symbol=target_symbol,
                                trading_symbol=trading_symbol,
                                algorithm_context=self
                            )
                        )

                    data["order_books"] = order_books
                else:
                    data["order_book"] = data_provider.provide_order_book(
                        target_symbol=target_symbol,
                        trading_symbol=trading_symbol,
                        algorithm_context=self
                    )
            elif TradingDataTypes.RAW.equals(trading_data_type):
                data["raw_data"] = data_provider.provide_raw_data(
                    target_symbol=target_symbol,
                    trading_symbol=trading_symbol,
                    algorithm_context=self
                )

        if trading_data_types is not None:
            if [trading_data_type for trading_data_type in trading_data_types
                    if TradingDataTypes.TICKER.equals(trading_data_type)]:

                if target_symbols is not None:
                    tickers = []

                    for target_symbol in target_symbols:
                        tickers.append(
                            data_provider.provide_ticker(
                                target_symbol=target_symbol,
                                trading_symbol=trading_symbol,
                                algorithm_context=self
                            )
                        )

                    data["tickers"] = tickers

                data["ticker"] = data_provider.provide_ticker(
                    target_symbol=target_symbol,
                    trading_symbol=trading_symbol,
                    algorithm_context=self
                )

            if [trading_data_type for trading_data_type in trading_data_types
                    if TradingDataTypes.ORDER_BOOK.equals(trading_data_type)]:

                if target_symbols is not None:
                    order_books = []

                    for target_symbol in target_symbols:
                        order_books.append(
                            data_provider.provide_order_book(
                                target_symbol=target_symbol,
                                trading_symbol=trading_symbol,
                                algorithm_context=self
                            )
                        )

                    data["order_books"] = order_books

                data["order_book"] = data_provider.provide_order_book(
                    target_symbol=target_symbol,
                    trading_symbol=trading_symbol,
                    algorithm_context=self
                )

            if [trading_data_type for trading_data_type in trading_data_types
                    if TradingDataTypes.RAW.equals(trading_data_type)]:

                if target_symbols is not None:
                    raw_data = []

                    for target_symbol in target_symbols:
                        raw_data.append(
                            data_provider.provide_raw_data(
                                target_symbol=target_symbol,
                                trading_symbol=trading_symbol,
                                algorithm_context=self
                            )
                        )

                    data["raw_data"] = raw_data

                if target_symbol is not None:
                    data["raw_data"] = data_provider.provide_raw_data(
                        target_symbol=target_symbol,
                        trading_symbol=trading_symbol,
                        algorithm_context=self
                    )

                else:
                    data["raw_data"] = data_provider.provide_raw_data(
                        trading_symbol=trading_symbol,
                        algorithm_context=self
                    )

        return data

    def get_unallocated_size(self, identifier):
        portfolio_manager = self.get_portfolio_manager(identifier)
        return portfolio_manager.unallocated

    def reset(self):
        self._workers = []
        self._running_workers = []
        self._data_providers = {}
        self._order_executors = {}
        self._portfolio_managers = {}
        self._market_services = {}
        self._initializer = None
        self._initialized = False

    def get_orders(
            self, identifier=None, symbol: str = None, status=None, lazy=False
    ) -> List[Order]:
        portfolio_manager = self.get_portfolio_manager(identifier)
        return portfolio_manager.get_orders(symbol, status, lazy)

    def get_positions(
            self, identifier=None, symbol: str = None, lazy=False
    ) -> List[Position]:
        portfolio_manager = self.get_portfolio_manager(identifier)
        return portfolio_manager.get_positions(symbol, lazy)

    def get_portfolio(self, identifier=None) -> Portfolio:
        portfolio_manager = self.get_portfolio_manager(identifier)
        return portfolio_manager.get_portfolio()



from enum import Enum


class TimeUnit(Enum):
    SECONDS = "SECONDS"
    MINUTE = "MINUTE"
    HOUR = "HOUR"

    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):

            for entry in TimeUnit:

                if value.upper() == entry.value:
                    return entry

        raise ValueError(
            "Could not convert value to time unit"
        )

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value
        else:
            return TimeUnit.from_string(other) == self



import random
import string


def random_string(n, spaces: bool = False):

    if spaces:
        return ''.join(
            random.choice(string.ascii_lowercase + ' ') for _ in range(n)
        )

    return ''.join(random.choice(string.ascii_lowercase) for _ in range(n))




import logging

import sqlalchemy
from flask import Blueprint, request

from investing_algorithm_framework import current_app
from investing_algorithm_framework import Order, Position, \
    Portfolio, OrderSide, db
from investing_algorithm_framework.configuration.constants import \
    TARGET_SYMBOL_QUERY_PARAM, TRADING_SYMBOL_QUERY_PARAM, \
    IDENTIFIER_QUERY_PARAM, ORDER_SIDE_QUERY_PARAM, STATUS_QUERY_PARAM, \
    POSITION_SYMBOL_QUERY_PARAM
from investing_algorithm_framework.schemas import OrderSerializer
from investing_algorithm_framework.views.utils import normalize_query, \
    create_paginated_response, get_query_param
from investing_algorithm_framework import OperationalException, ApiException

logger = logging.getLogger(__name__)

blueprint = Blueprint("order-views", __name__)


def apply_order_query_parameters(query_set):
    query_params = normalize_query(request.args)

    if IDENTIFIER_QUERY_PARAM in query_params:
        portfolio = Portfolio.query\
            .filter_by(identifier=query_params[IDENTIFIER_QUERY_PARAM])\
            .first()

        if portfolio is None:
            return db.session.query(Order).filter(sqlalchemy.sql.false())

        position_ids = portfolio.positions.with_entities(Position.id)
        query_set = query_set.filter(Order.position_id.in_(position_ids))

    if POSITION_SYMBOL_QUERY_PARAM in query_params:
        position_ids = Position.query\
            .filter_by(symbol=query_params[POSITION_SYMBOL_QUERY_PARAM])\
            .with_entities(Position.id)

        query_set = query_set.filter(Order.position_id.in_(position_ids))

    if TARGET_SYMBOL_QUERY_PARAM in query_params:
        query_set = query_set.filter_by(
            target_symbol=query_params[TARGET_SYMBOL_QUERY_PARAM]
        )

    if TRADING_SYMBOL_QUERY_PARAM in query_params:
        query_set = query_set.filter_by(
            trading_symbol=query_params[TRADING_SYMBOL_QUERY_PARAM]
        )

    if ORDER_SIDE_QUERY_PARAM in query_params:
        query_set = query_set.filter_by(
            order_side=OrderSide.from_string(
                query_params[ORDER_SIDE_QUERY_PARAM]
            ).value
        )

    if STATUS_QUERY_PARAM in query_params:
        query_set = query_set\
            .filter_by(status=query_params[STATUS_QUERY_PARAM])

    return query_set


@blueprint.route("/api/orders", methods=["GET"])
def list_orders():
    """
    View for listing of the orders of the algorithm. This view will list all
    the orders of all portfolios of your algorithm.

    You can provide to this view the following query params:
        - target_symbol: the symbol that is traded in the order.
        - trading_symbol: the symbol that is traded with in the order
        - order_side: the order side of the order (BUY/SELL)

    The response in the view is paginated.
    """

    try:
        identifier = get_query_param(IDENTIFIER_QUERY_PARAM, request.args)
        portfolio = current_app.algorithm.get_portfolio(identifier)
        status = get_query_param(STATUS_QUERY_PARAM, request.args)
        symbol = get_query_param(TARGET_SYMBOL_QUERY_PARAM, request.args)
        side = get_query_param(ORDER_SIDE_QUERY_PARAM, request.args)

        orders = portfolio\
            .get_orders(
                status=status,
                side=side,
                target_symbol=symbol
            )

        # Create serializer
        serializer = OrderSerializer()
    except OperationalException as e:
        raise ApiException(e.error_message)

    # Paginate query
    return create_paginated_response(
        orders,
        serializer,
        get_query_param("page", request.args, 1),
        get_query_param("per_page", request.args, 20)
    ), 200



from investing_algorithm_framework.core.models import PerformanceMetric, \
    TimeFrame
from investing_algorithm_framework.core.performance import \
    SnapshotAssetPriceCollection


class PerformanceService:

    @staticmethod
    def of_metric(portfolio, metric: PerformanceMetric, time_frame: TimeFrame):
        # Create SnapshotAssetPriceCollection object for iteration
        snapshot_asset_price_collection = SnapshotAssetPriceCollection(
            portfolio=portfolio,
            time_frame=time_frame,
        )

        if PerformanceMetric.OVERALL_PERFORMANCE.equals(metric):
            return PerformanceService().generate_overall_performance(
                snapshot_asset_price_collection
            )
        elif PerformanceMetric.DELTA.equals(metric):
            return PerformanceService().generate_delta(
                snapshot_asset_price_collection
            )

        raise NotImplementedError("Performance metric not supported")

    def generate_delta(self, snapshot_asset_price_collection):
        """
            Performance calculation with Time Weighted Return
            TWR = [(1 + HP_1) * (1 + HP_2) * ... * (1 + HP_n)] - 1

            Where
                TWR = Time weighted return
                n = Number of sub-periods
                HP = (Total Value - (Previous Total Value + Cash Flow))
                        / (Previous Total Value + Cash Flow)
                HP_n = Return for sub-period n
        """
        points = []
        previous_snapshot = None
        previous_asset_prices = None
        previous_point = 1.0
        delta = 1

        points.append(previous_point)

        for date, snapshot, asset_prices in snapshot_asset_price_collection:
            points.append(1 + self.get_hp(
                snapshot,
                previous_snapshot,
                asset_prices,
                previous_asset_prices
            ))

            previous_asset_prices = asset_prices
            previous_snapshot = snapshot

        for point in points:
            delta = delta * point

        return delta - 1

    def generate_overall_performance(self, snapshot_asset_price_collection):
        """
            Performance calculation with Time Weighted Return
            TWR = [(1 + HP_1) * (1 + HP_2) * ... * (1 + HP_n)] - 1

            Where
                TWR = Time weighted return
                n = Number of sub-periods
                HP = (Total Value - (Previous Total Value + Cash Flow))
                        / (Previous Total Value + Cash Flow)
                HP_n = Return for sub-period n
        """
        points = []
        previous_snapshot = None
        previous_asset_prices = None
        previous_point = 1.0

        for date, snapshot, asset_prices in snapshot_asset_price_collection:
            hp_value = 1 + self.get_hp(
                snapshot,
                previous_snapshot,
                asset_prices,
                previous_asset_prices
            )

            points.append(
                {
                    "datetime": date,
                    "value": (previous_point * hp_value) - 1
                }
            )

            previous_asset_prices = asset_prices
            previous_snapshot = snapshot
            previous_point = (previous_point * hp_value)

        return points

    def get_hp(
        self,
        snapshot,
        previous_snapshot,
        current_asset_prices,
        previous_asset_prices
    ):
        if snapshot is not None and previous_snapshot is not None:
            return (
                snapshot.get_total_value(current_asset_prices) -
                (previous_snapshot.get_total_value(previous_asset_prices)
                 + snapshot.cash_flow)) \
               / (previous_snapshot.get_total_value(previous_asset_prices)
                  + snapshot.cash_flow)

        return 0



from investing_algorithm_framework.utils.random import random_string
from investing_algorithm_framework.utils.stoppable_thread import \
    StoppableThread
from investing_algorithm_framework.utils.synchronized import synchronized

__all__ = ['synchronized', 'StoppableThread', 'random_string']



import inspect
import os
import shutil
import pathlib
from distutils.sysconfig import get_python_lib

from flask import Flask

from investing_algorithm_framework.configuration import Config, create_app, \
    setup_config, setup_database, setup_logging
from investing_algorithm_framework.context import Singleton
from investing_algorithm_framework.core.context import algorithm
from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import create_all_tables, \
    initialize_db
from investing_algorithm_framework.extensions import scheduler
from investing_algorithm_framework.configuration.constants import \
    RESOURCES_DIRECTORY, SQLALCHEMY_DATABASE_URI, DATABASE_DIRECTORY_PATH, \
    DATABASE_NAME, DATABASE_CONFIG, LOG_LEVEL


class App(metaclass=Singleton):
    _algorithm = algorithm
    _flask_app: Flask = None
    _configured: bool = False
    _database_configured: bool = False
    _started = False
    _config = None
    _resource_directory = None
    _blueprints = []

    def __init__(
        self, resources_directory: str = None, config=None, arg=None
    ):
        if resources_directory is not None:
            self._resource_directory = resources_directory

        if config is not None:
            self._config = config

    def initialize(
            self, resources_directory: str = None, config=None, arg=None
    ):
        if not self.started:

            if resources_directory is not None:
                self._resource_directory = resources_directory

            if config is not None:
                self._config = config

    def _initialize_algorithm(self):
        self._algorithm.initialize(config=self.config)

    def _initialize_config(self, config=None):

        if not self._configured:

            if config is not None:
                self._config = config

            if self._config is None:
                raise OperationalException("No config object set")

            if inspect.isclass(self._config) \
                    and issubclass(self._config, Config):
                self._config = self._config()
            elif type(self._config) is dict:
                self._config = Config.from_dict(self._config)
            else:
                raise OperationalException("Config object not supported")

            if self._resource_directory is not None:
                self._config[RESOURCES_DIRECTORY] = self._resource_directory

            if RESOURCES_DIRECTORY not in self._config \
                    or self._config[RESOURCES_DIRECTORY] is None:
                raise OperationalException("Resource directory not specified")

            self._configured = True
            setup_logging(self.config.get(LOG_LEVEL, "INFO"))

    def _initialize_flask_app(self):

        if self._flask_app is None:
            self._flask_app = create_app()

    def _initialize_flask_sql_alchemy(self):

        if self._configured and self._database_configured:
            initialize_db(self._flask_app)
            create_all_tables()

    def _initialize_flask_config(self):

        if self._configured:
            setup_config(self._flask_app, self._config)

    def _initialize_database(self):

        if self._configured and not self._database_configured:
            setup_database(self.config)

            if self.config[DATABASE_CONFIG][DATABASE_DIRECTORY_PATH] is None:
                raise OperationalException(
                    f"{DATABASE_DIRECTORY_PATH} is not set in config"
                )

            if self.config[DATABASE_CONFIG][DATABASE_NAME] is None:
                raise OperationalException(
                    f"{DATABASE_NAME} is not set in config"
                )

            if self.config[SQLALCHEMY_DATABASE_URI] is None:
                raise OperationalException(
                    f"{SQLALCHEMY_DATABASE_URI} is not set in config"
                )

            self._database_configured = True

    def _initialize_management_commands(self):
        # Copy the template manage.py file to the resource directory of the
        # algorithm
        management_commands_template = os.path.join(
            get_python_lib(),
            "investing_algorithm_framework/templates/manage.py"
        )

        destination = os.path.join(self._resource_directory, "manage.py")

        if not os.path.exists(destination):
            shutil.copy(management_commands_template, destination)

    def register_blueprint(self, blueprint):
        self._blueprints.append(blueprint)

    def _initialize_blueprints(self):

        for blueprint in self._blueprints:
            self._flask_app.register_blueprint(blueprint)

    def start(self):

        # Setup config if it is not set
        if self._config is None:
            self._config = Config

        self._initialize_flask_app()
        self._initialize_blueprints()
        self._initialize_config()
        self._initialize_database()
        self._initialize_flask_config()
        self._initialize_flask_sql_alchemy()
        self._initialize_algorithm()
        self._initialize_management_commands()

        self.start_scheduler()
        self.start_algorithm()

        # Start the app
        self._flask_app.run(
            debug=True,
            threaded=True,
            use_reloader=False
        )

    def start_scheduler(self):

        # Initialize the schedulers
        if not scheduler.running:
            scheduler.init_app(self._flask_app)
            scheduler.start()

    def start_algorithm(self):

        if not scheduler.running:
            raise OperationalException(
                "Could not start algorithm because the scheduler "
                "is not running"
            )

        # Start the algorithm
        self._algorithm.start()

    def stop_algorithm(self):

        if self._algorithm.running:
            self._algorithm.stop()

    @property
    def algorithm(self):
        return self._algorithm

    @property
    def started(self):
        return self._started

    @property
    def config(self):
        return self._config

    def reset(self):
        self._configured = False
        self._database_configured: bool = False
        self._started = False
        scheduler.remove_all_jobs()
        self.algorithm.reset()

    def add_initializer(self, initializer):
        self.algorithm.add_initializer(initializer)

    def add_order_executor(self, order_executor):
        self.algorithm.add_order_executor(order_executor)

    def add_data_provider(self, data_provider):
        self.algorithm.add_data_provider(data_provider)

    def add_portfolio_manager(self, portfolio_manager):
        self.algorithm.add_portfolio_manager(portfolio_manager)



from datetime import datetime
from typing import List

from investing_algorithm_framework.core.models.model_extension import \
    SQLAlchemyModelExtension


class OrderBook(SQLAlchemyModelExtension):

    def __init__(self, symbol, bids: List, asks: List, creation_date=None):
        self.symbol = symbol
        self.bids = bids
        self.asks = asks

        if creation_date is None:
            self.creation_date = datetime.now()

    def get_bids(self):
        return self.bids

    def get_asks(self):
        return self.asks

    def get_symbol(self):
        return self.symbol

    def to_dict(self):

        return {
            "symbol": self.symbol,
            "bids": self.bids,
            "asks": self.asks
        }

    @staticmethod
    def from_dict(data):
        return OrderBook(
            data.get("symbol"),
            data.get("bids"),
            data.get("asks"),
            data.get("creation_date")
        )

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []
        at_least_one_attached_attribute = False

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')
            at_least_one_attached_attribute = True

        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"

        return f"<{self.__class__.__name__} {id(self)}>"

    def __repr__(self):
        return self.repr(
            symbol=self.symbol,
            bids=self.bids,
            asks=self.asks
        )
